import itertools

from plado.datalog.evaluator import Atom, Clause, Constant, DatalogEngine
from plado.semantics.task import Action, DelEffect, State, Task
from plado.utils import Float, is_less, is_less_equal, is_zero

Successors = list[tuple[State, Float]]


class ActionSuccessorGenerator:
    def __init__(self, task: Task, action: Action):
        self.task: Task = task
        self.action: Action = action
        program = self.task.create_datalog_program()
        param_relation: int = program.num_relations()
        for _ in range(self.action.parameters):
            program.add_relation(1)
        self.effect_relation: int = program.num_relations()
        self.always_effects: list[int] = []
        counter = 0
        for eff in self.action.effect.effects:
            for _, out in eff.outcomes:
                for atom_eff in out:
                    assert atom_eff.idx == counter
                    counter += 1
                    r = program.add_relation(
                        len(eff.parameters) + len(atom_eff.parameters)
                    )
                    assert r == self.effect_relation + atom_eff.idx
                    clause = Clause(
                        Atom(
                            r,
                            [
                                Constant(var_id, True)
                                for var_id in itertools.chain(
                                    eff.parameters, atom_eff.parameters
                                )
                            ],
                        ),
                        [],
                        [],
                        [],
                    )
                    atom_eff.condition.to_datalog(len(self.task.predicates), clause)
                    for param in atom_eff.condition.get_variables():
                        if param < self.action.parameters:
                            clause.pos_body = clause.pos_body + tuple(
                                [Atom(param_relation + param, [Constant(param, True)])]
                            )
                        else:
                            break
                    if clause.is_trivial():
                        self.always_effects.append(atom_eff.idx)
                    else:
                        program.add_clause(clause)
        if len(self.always_effects) != self.action.effect.num_atomic_effects:
            self.evaluator: DatalogEngine = DatalogEngine(
                program, len(self.task.objects)
            )
        else:
            self.evaluator = None

    def _merge(self, effs0, effs1):
        if len(effs1) == 0:
            return effs0
        p1sum = sum((p for p, _ in effs1))
        assert is_less_equal(p1sum, 1), f"successor probabilities sum up to {p1sum} > 1"
        if is_less(p1sum, 1):
            effs1.append((1 - p1sum, []))
        if len(effs0) == 0:
            return effs1
        res = []
        for p0, e0 in effs0:
            for p1, e1 in effs1:
                res.append((p0 * p1, e0 + e1))
        return res

    def _collect_outcomes_unconditional(self, state: State, grounding: tuple[int]):
        result = []
        for peff in self.action.effect.effects:
            assert len(peff.parameters) == 0
            outcomes = []
            for prob, out in peff.outcomes:
                try:
                    pval = prob.evaluate(grounding, state)
                    if is_zero(pval):
                        continue
                except KeyError:
                    continue
                effs = []
                for eff in out:
                    assert len(eff.parameters) == 0
                    effs.append(eff.effect.instantiate(grounding, state))
                outcomes.append((pval, effs))
            result = self._merge(result, outcomes)
        return result

    def _collect_outcomes(self, state: State, grounding: tuple[int]):
        assert self.evaluator is not None
        atoms, fluents = self.task.prepare_for_query(state.atoms, state.fluents)
        for obj in grounding:
            atoms.append(set([tuple([obj])]))
        atoms.extend((set() for _ in range(self.action.effect.num_atomic_effects)))
        model = self.evaluator(atoms, fluents)
        result = []
        k = 0
        for peff in self.action.effect.effects:
            # outcomes parameterized by universal var assignment (i.e., treating
            # different assignments as different probabilistic effects)
            outcomes = {}
            for prob, out in peff.outcomes:
                # unparameterized effects (empty if the probabilistic effect is
                # parameterized)
                certain_effs = []
                # parameterized effects where either the probabilistic effect is
                # wrapped in a universal effect, or the atomic effect is
                param_effs = {}
                for eff in out:
                    # check if unparameterized
                    if (
                        k < len(self.always_effects)
                        and self.always_effects[k] == eff.idx
                    ):
                        # out.effect has no parameters other than of the action
                        certain_effs.append(eff.effect.instantiate(grounding, state))
                        k += 1
                    else:
                        # iterate over all universal quantification variables
                        # where the condition is satisfied
                        for params in model[self.effect_relation + eff.idx]:
                            # append the universal variable instantiations to the
                            # grounding
                            extgrnd = grounding + params
                            # project out universal variables of the
                            # probabilistic effect -> the instantiated outcome
                            # goes into the outcome list of the corresponding
                            # instantiated probabilistic effect
                            aparams = params[: len(peff.parameters)]
                            if aparams not in param_effs:
                                try:
                                    # probabilistic effect instantiated with
                                    # aparams for the first time -> evaluate the
                                    # probability
                                    pval = prob.evaluate(extgrnd, state.fluents)
                                    param_effs[aparams] = [pval, []]
                                except KeyError:
                                    param_effs[aparams] = [None, []]
                            # add the outcome to the effects of the
                            # corresponding probabilistic action instance
                            param_effs[aparams][1].append(
                                eff.effect.instantiate(extgrnd, state)
                            )
                if len(certain_effs) != 0:
                    # if there are certain effs, the probabilistic effect is not
                    # instantiated -> aparams = (,) is an instance
                    aparams = tuple([])
                    if aparams not in outcomes:
                        try:
                            pval = prob.evaluate(grounding, state.fluents)
                            if not is_zero(pval):
                                outcomes.setdefault(aparams, []).append(
                                    (pval, certain_effs)
                                )
                        except KeyError:
                            pass
                for aparams, (pval, effs) in param_effs.items():
                    if pval is not None and not is_zero(pval):
                        effs.extend(certain_effs)
                        outcomes.setdefault(aparams, []).append((pval, effs))
            for outs in outcomes.values():
                result = self._merge(result, outs)
        return result

    def __call__(self, state: State, grounding: tuple[int]) -> Successors:
        if self.evaluator is None:
            outcomes = self._collect_outcomes_unconditional(state, grounding)
        else:
            outcomes = self._collect_outcomes(state, grounding)
        result = []
        for prob, effs in outcomes:
            succ = state.duplicate()
            i = 0
            for j in range(len(effs)):
                if isinstance(effs[j], DelEffect):
                    effs[j].apply(succ)
                else:
                    if i != j:
                        effs[i] = effs[j]
                    i += 1
            for j in range(i):
                effs[j].apply(succ)
            result.append((succ, prob))
        return result


class SuccessorGenerator:
    def __init__(self, task: Task):
        self.gens: list[ActionSuccessorGenerator] = [
            ActionSuccessorGenerator(task, action) for action in task.actions
        ]

    def __call__(
        self, state: State, ground_action: tuple[int, tuple[int]]
    ) -> Successors:
        return self.gens[ground_action[0]](state, ground_action[1])



import re
from enum import Enum, auto


class Category(Enum):
    LBRACK = auto()
    RBRACK = auto()
    MINUS = auto()
    DEFINE = auto()
    NAME = auto()
    VARIABLE = auto()
    DOMAIN_DEF = auto()
    PROBLEM_DEF = auto()
    DOMAIN = auto()
    PROBLEM = auto()
    REQUIREMENTS = auto()
    TYPES = auto()
    CONSTANTS = auto()
    PREDICATES = auto()
    FUNCTIONS = auto()
    ACTION = auto()
    PARAMETERS = auto()
    EFFECT = auto()
    PRECONDITION = auto()
    OBJECTS = auto()
    INIT = auto()
    GOAL = auto()
    GOAL_REWARD = auto()
    METRIC = auto()
    MINIMIZE = auto()
    MAXIMIZE = auto()
    EITHER = auto()
    NUMBER = auto()
    AND = auto()
    NOT = auto()
    OR = auto()
    IMPLY = auto()
    EXISTS = auto()
    FORALL = auto()
    WHEN = auto()
    PROBABILISTIC = auto()
    ASSIGN = auto()
    SCALE_UP = auto()
    SCALE_DOWN = auto()
    INCREASE = auto()
    DECREASE = auto()
    EQUAL = auto()
    LT = auto()
    LE = auto()
    GT = auto()
    GE = auto()
    PLUS = auto()
    TIMES = auto()
    DIVIDE = auto()
    CONSTANT = auto()
    REQUIREMENT_CONSTANT = auto()
    DERIVED = auto()


NAME_RE = re.compile(r"[a-z][\w\-_]*", re.IGNORECASE)

VARIABLE_RE = re.compile(r"\?[a-z][\w\-_]*", re.IGNORECASE)

CONSTANT_RE = re.compile(r"-?\d*\.?\d+(e-?\d+)?(/?\d\+)?")

WS_RE = re.compile(r"(\s+|\)|;)")

REQUIREMENTS_RE = re.compile(
    ":(strips|typing|equality|negative-preconditions|disjunctive-preconditions|existential-preconditions|universal-preconditions|quantified-preconditions|conditional-effects|probabilistic-effects|rewards|fluents|adl|mdp|derived-predicate|action-costs)"
)


class Token:
    def __init__(self, cat, lno, cno, tok):
        # category
        self.cat = cat
        # line number
        self.lno = lno
        # column number
        self.cno = cno
        # token string
        self.tok = tok

    def __repr__(self):
        return "<Token %r %r @%d:%d>" % (self.cat, self.tok, self.lno, self.cno)


def parse_token(word, lno, cno):
    lword = word.lower()
    for match, match_cat in [
        ("-", Category.MINUS),
        ("define", Category.DEFINE),
        ("domain", Category.DOMAIN_DEF),
        ("problem", Category.PROBLEM_DEF),
        (":problem", Category.PROBLEM),
        (":domain", Category.DOMAIN),
        (":requirements", Category.REQUIREMENTS),
        (":types", Category.TYPES),
        (":constants", Category.CONSTANTS),
        (":predicates", Category.PREDICATES),
        (":functions", Category.FUNCTIONS),
        (":derived", Category.DERIVED),
        (":action", Category.ACTION),
        (":parameters", Category.PARAMETERS),
        (":effect", Category.EFFECT),
        (":precondition", Category.PRECONDITION),
        (":objects", Category.OBJECTS),
        (":init", Category.INIT),
        (":goal", Category.GOAL),
        (":goal-reward", Category.GOAL_REWARD),
        (":metric", Category.METRIC),
        ("minimize", Category.MINIMIZE),
        ("maximize", Category.MAXIMIZE),
        ("either", Category.EITHER),
        ("and", Category.AND),
        ("not", Category.NOT),
        ("or", Category.OR),
        ("imply", Category.IMPLY),
        ("exists", Category.EXISTS),
        ("forall", Category.FORALL),
        ("when", Category.WHEN),
        ("probabilistic", Category.PROBABILISTIC),
        ("assign", Category.ASSIGN),
        ("scale-up", Category.SCALE_UP),
        ("scale-down", Category.SCALE_DOWN),
        ("increase", Category.INCREASE),
        ("decrease", Category.DECREASE),
        ("=", Category.EQUAL),
        ("<", Category.LT),
        ("<=", Category.LE),
        (">", Category.GT),
        (">=", Category.GE),
        ("+", Category.PLUS),
        ("*", Category.TIMES),
        ("/", Category.DIVIDE),
    ]:
        if word == match:
            return Token(match_cat, lno, cno, word)
    if REQUIREMENTS_RE.match(lword):
        return Token(Category.REQUIREMENT_CONSTANT, lno, cno, word)
    if NAME_RE.match(lword):
        return Token(Category.NAME, lno, cno, word)
    if VARIABLE_RE.match(lword):
        return Token(Category.VARIABLE, lno, cno, word)
    if CONSTANT_RE.match(lword):
        return Token(Category.CONSTANT, lno, cno, word)
    raise ValueError(f"Cannot parse token {word} at line {lno} and column {cno}")


def tokenize(pddl_content):
    pddl_content = pddl_content.lower().replace("?", " ?").replace("(", " (")
    # current line number
    lno = 1
    # position of first character of current line
    line_start = 0
    # iterate over all lines
    while line_start < len(pddl_content):
        # find end position of current line
        line_end = pddl_content.find("\n", line_start)
        # if no new line, set to end of string
        if line_end < 0:
            line_end = len(pddl_content)
        # check if line is empty
        if line_start < line_end:
            # column index
            cno = 1
            # process word by word
            wstart = line_start
            while wstart < line_end:
                # special case handline for brackets
                if pddl_content[wstart] == "(":
                    yield Token(Category.LBRACK, lno, cno, "(")
                    wnext = wstart + 1
                elif pddl_content[wstart] == ")":
                    yield Token(Category.RBRACK, lno, cno, ")")
                    wnext = wstart + 1
                # handle comment
                elif pddl_content[wstart] == ";":
                    wnext = line_end
                else:
                    # find end of word
                    wend = WS_RE.search(pddl_content, pos=wstart)
                    # set to end of line if no whitespace character found
                    if wend is None:
                        wend = line_end
                        wnext = line_end
                    else:
                        wnext = wend.end() - (
                            int(pddl_content[wend.end() - 1] in [")", ";"])
                        )
                        wend = wend.start()
                    # parse token if word is not empty
                    if wstart < wend:
                        word = pddl_content[wstart:wend]
                        yield parse_token(word, lno, cno)
                # update column position
                cno += wnext - wstart
                # continue after last whitespace
                wstart = wnext
        # progress to next line
        line_start = line_end + 1
        lno += 1



from plado import pddl
from plado.utils import reachability_closure


def get_type_closure(domain: pddl.Domain) -> dict[str, list[str]]:
    hierarchy = {t.name: [t.parent_type_name] for t in domain.types}
    hierarchy["object"] = []
    root_name = "@root@"
    assert root_name not in hierarchy
    closure = reachability_closure(
        root_name, lambda name: hierarchy.get(name, hierarchy.keys())
    )
    del closure[root_name]
    return closure



from collections.abc import Iterable

import plado.datalog.program as datalog
from plado.datalog.evaluator import DatalogEngine
from plado.semantics.task import AtomsStore, FluentsStore, State, Task

GroundActionRef = tuple[int, tuple[int]]


class ApplicableActionsGenerator:
    def __init__(self, task: Task):
        program = task.create_datalog_program()
        self.task: Task = task
        self.num_actions = len(task.actions)
        self.offset = program.num_relations()
        for action in task.actions:
            rel_id = program.add_relation(action.parameters)
            clause = datalog.Clause(
                datalog.Atom(
                    rel_id,
                    (datalog.Constant(var, True) for var in range(action.parameters)),
                ),
                [],
                [],
                [],
            )
            action.precondition.to_datalog(len(task.predicates), clause)
            program.add_clause(clause)
        self.datalog = DatalogEngine(program, len(task.objects))

    def _prepare_for_query(self, state: State) -> tuple[AtomsStore, FluentsStore]:
        atoms, fluents = self.task.prepare_for_query(state.atoms, state.fluents)
        atoms.extend((set() for _ in range(self.num_actions)))
        return atoms, fluents

    def __call__(self, state: State) -> Iterable[GroundActionRef]:
        atoms, fluents = self._prepare_for_query(state)
        model = self.datalog(atoms, fluents)
        for i in range(self.offset, len(model)):
            for args in model[i]:
                yield (i - self.offset, args)



import sys
from collections.abc import Iterable

from plado import pddl
from plado.pddl_utils import (
    get_type_closure,
    visit_all_conditions,
    visit_all_effects,
    visit_all_expressions,
    visit_all_expressions_in_condition,
)
from plado.utils.graph import tarjan


def _error(file, message):
    if file:
        print("[!]", message, file=file)


def _check_unique_names(
    elements: Iterable[
        pddl.Argument
        | pddl.ArgumentDefinition
        | pddl.Type
        | pddl.Predicate
        | pddl.Function
        | pddl.Action
    ],
    file=sys.stderr,
    name="",
) -> bool:
    """
    Checks that all argument names are unique.
    """
    duplicates = False
    names = set()
    for t in elements:
        if t.name in names:
            duplicates = True
            _error(file, f"Multiple {name} with name {t.name}")
        names.add(t.name)
    return not duplicates


def unique_object_names(
    objects: Iterable[pddl.ArgumentDefinition], file=sys.stderr
) -> bool:
    """
    Checks that all object have a unique name.
    """
    return _check_unique_names(objects, file, "objects")


def unique_type_names(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that all types have a unique name.
    """
    return _check_unique_names(domain.types, file, "types")


def unique_action_names(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that all actions have a unique name.
    """
    return _check_unique_names(domain.actions, file, "actions")


def unique_predicates(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that all predicates have a unique name.
    """
    return _check_unique_names(domain.predicates, file, "predicates")


def unique_functions(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that all functions have a unique name.
    """
    return _check_unique_names(domain.functions, file, "functions")


def no_reserved_functions(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that no reserved function name is used (i.e.,
    total-cost, reward, and goal-reward). If they are defined, they are
    expected to have 0-arity.
    """
    error = False
    for fname in ["total-cost", "reward", "goal-reward"]:
        for f in domain.functions:
            if f.name == fname:
                if len(f.parameters) != 0:
                    error = True
                    _error(file, "{fname} is a reserved function name")
                break
    return error


def unique_variable_names(
    domain: pddl.Domain, problem: pddl.Problem | None, file=sys.stderr
) -> bool:
    """
    Checks that variable names are unique at every place where variables might appear,
    i.e.,  in predicate and function definitions, in action schema parameters, in
    existential and universal quantification, and universal effects.
    """

    def check_unique(parameters: tuple[pddl.ArgumentDefinition]) -> bool:
        return len(set((p.name for p in parameters))) != len(parameters)

    class UniqueVarChecker(pddl.RecursiveBooleanExpressionVisitor):
        def __init__(self):
            self.error = False

        def visit_generic(self, formula: pddl.BooleanExpression):
            raise ValueError()

        def visit_atomic(self, formula: pddl.BooleanExpression) -> bool:
            return False

        def visit_quantification(self, formula: pddl.Quantification) -> bool:
            formula.sub_formula.traverse(self)
            self.error = self.error or check_unique(formula.parameters)
            return False

    class UniqueVarCheckerEffect(pddl.RecursiveActionEffectVisitor):
        def __init__(self):
            self.error = False

        def visit_atomic(self, effect: pddl.ActionEffect) -> bool:
            return False

        def visit_universal_effect(self, effect: pddl.UniversalEffect) -> bool:
            effect.effect.traverse(self)
            self.error = self.error or check_unique(effect.parameters)
            return False

    def check_action_pre(action: pddl.Action) -> bool:
        visitor = UniqueVarChecker()
        action.precondition.traverse(visitor)
        return visitor.error

    def check_action_effect(action: pddl.Action) -> bool:
        visitor = UniqueVarCheckerEffect()
        action.effect.traverse(visitor)
        return visitor.error

    def check_derived_predicate(predicate: pddl.DerivedPredicate) -> bool:
        return check_unique(predicate.predicate.parameters)

    def proj(x):
        return x.parameters

    preds = list(filter(check_unique, map(proj, domain.predicates)))
    funs = list(filter(check_unique, map(proj, domain.functions)))
    acts_params = list(filter(check_unique, map(proj, domain.actions)))
    dpreds = list(filter(check_derived_predicate, domain.derived_predicates))
    acts_pre = list(filter(check_action_pre, domain.actions))
    acts_eff = list(filter(check_action_effect, domain.actions))
    goal = False
    if problem:
        visitor = UniqueVarChecker()
        problem.goal.traverse(visitor)
        goal = visitor.error
    if file:
        map(
            lambda p: _error(
                file,
                "Predicate %r has multiple parameters with the same name." % p.name,
            ),
            preds,
        )
        map(
            lambda p: _error(
                file, "Function %r has multiple parameters with the same name." % p.name
            ),
            funs,
        )
        map(
            lambda p: _error(
                file, "Action %r has multiple parameters with the same name." % p.name
            ),
            acts_params,
        )
        map(
            lambda p: _error(
                file,
                "A quantifier in action %r has multiple parameters with the same name."
                % p.name,
            ),
            acts_pre,
        )
        map(
            lambda p: _error(
                file,
                "A universal effect of action %r has multiple parameters with the same"
                " name."
                % p.name,
            ),
            acts_eff,
        )
        map(
            lambda p: _error(
                file,
                "Derived predicate %r has multiple parameters with the same name."
                % p.name,
            ),
            dpreds,
        )
        # if goal:
        #     _error(file, "A quantifier in the goal has multiple parameters with the same name."))
    return (
        len(preds)
        + len(funs)
        + len(acts_params)
        + len(acts_pre)
        + len(acts_eff)
        + len(dpreds)
        + int(goal)
        == 0
    )


def type_hierarchy(domain: pddl.Domain, file=sys.stderr) -> bool:
    """
    Checks that all referenced type names in the type hierarchy are defined
    and that the induced type inheritance graph is acyclic.
    """
    hierarchy = {}
    for t in domain.types:
        hierarchy[t.name] = [t.parent_type_name]
    if "object" in hierarchy:
        if hierarchy["object"][0] not in ("object",):
            if file:
                print(
                    "'object' type may not be a sub-type of any other type",
                    file=file,
                )
            return False
    hierarchy["object"] = []
    for t in hierarchy.values():
        if len(t) > 0 and t[0] is not None and t[0] not in hierarchy:
            if file:
                print(f"Type {t[0]} referenced but not defined", file=file)
            return False

    class CycleChecker:
        def __init__(self):
            self.has_cycles: bool = False

        def __call__(self, scc: list[str]) -> None:
            if len(scc) > 1:
                _error(
                    file,
                    "Circular type inheritence between types " + " ".join(sorted(scc)),
                )
            self.has_cycles = self.has_cycles or len(scc) > 1

    root_name = "@root@"
    assert root_name not in hierarchy
    on_scc = CycleChecker()
    tarjan(root_name, lambda name: hierarchy.get(name, hierarchy.keys()), on_scc)
    return not on_scc.has_cycles


def predicate_references(
    domain: pddl.Domain, problem: pddl.Problem | None, file=sys.stderr
) -> bool:
    """
    Checks that all atoms reference a defined predicate with matching arity, and that
    all function calls reference a defined function with matching arity.
    """
    type_closure = get_type_closure(domain)
    predicates = {"=": tuple(["object", "object"])}
    for p in domain.predicates:
        predicates[p.name] = p.parameters
    functions = {"total-cost": tuple(), "reward": tuple(), "goal-achieved": tuple()}
    for f in domain.functions:
        functions[f.name] = f.parameters

    def function_not_found(t: str, ctxt: str) -> None:
        if file:
            print("Function %r referenced but not defined%s" % (t, ctxt), file=file)

    def predicate_not_found(t: str, ctxt: str) -> None:
        if file:
            print("Predicate %r referenced but not defined%s" % (t, ctxt), file=file)

    def type_not_found(t: str, ctxt: str) -> None:
        if file:
            print("Type %r referenced but not defined%s" % (t, ctxt), file=file)

    def predicate_argument_mismatch(atom: pddl.Atom | pddl.AtomEffect, ctxt: str):
        if file:
            print(
                "Argument mismatch in %s. Predicate expects %d parameter(s).%s"
                % (atom, len(predicates[atom.name]), ctxt)
            )

    def function_argument_mismatch(atom: pddl.FunctionCall, ctxt: str):
        if file:
            print(
                "Argument mismatch in %s. Function expects %d parameter(s).%s"
                % (atom, len(functions[atom.name]), ctxt)
            )

    def check_atom(atom: pddl.Atom | pddl.AtomEffect) -> bool:
        if atom.name not in predicates:
            predicate_not_found(atom.name, "")
            return True
        if len(atom.arguments) != len(predicates[atom.name]):
            predicate_argument_mismatch(atom, "")
            return True
        return False

    def check_function(fn: pddl.Function) -> bool:
        if fn.name not in functions:
            function_not_found(fn.name, "")
            return True
        if len(fn.arguments) != len(functions[fn.name]):
            function_argument_mismatch(fn, "")
            return True
        return False

    class PredicateChecker(pddl.RecursiveBooleanExpressionVisitor):
        def __init__(self, ctxt: str = ""):
            self.ctxt = ctxt
            self.error: bool = False

        def visit_atomic(self, formula: pddl.BooleanExpression) -> bool:
            return False

        def visit_atom(self, formula: pddl.Atom) -> bool:
            self.error = check_atom(formula) or self.error
            return False

        def visit_quantification(self, formula: pddl.Quantification) -> bool:
            formula.sub_formula.traverse(self)
            for x in formula.parameters:
                if x.type_name not in type_closure:
                    self.error = True
                    type_not_found(x.type_name, self.ctxt)
            return False

    class FunctionChecker(pddl.RecursiveNumericExpressionVisitor):
        def __init__(self, ctxt: str = ""):
            self.ctxt = ctxt
            self.error = False

        def visit_atomic(self, expr: pddl.NumericExpression) -> bool:
            return False

        def visit_function_call(self, expr: pddl.NumericExpression) -> bool:
            self.error = check_function(expr) or self.error
            return False

    class EffectChecker(pddl.RecursiveActionEffectVisitor):
        def __init__(self, ctxt: str = ""):
            self.ctxt = ctxt
            self.error = False

        def visit_atomic(self, effect: pddl.ActionEffect) -> bool:
            return False

        def visit_atom_effect(self, effect: pddl.ActionEffect) -> bool:
            self.error = check_atom(effect) or self.error
            return False

        def visit_universal_effect(self, effect: pddl.ActionEffect) -> bool:
            effect.effect.traverse(self)
            for x in effect.parameters:
                if x.type_name not in type_closure:
                    self.error = True
                    type_not_found(x.type_name, self.ctxt)
            return False

    error = False
    pc = PredicateChecker()
    fc = FunctionChecker()
    ec = EffectChecker()

    visit_all_conditions(domain, pc)
    visit_all_expressions(domain, fc)
    visit_all_effects(domain.actions, ec)

    if problem:
        problem.goal.traverse(pc)
        visit_all_expressions_in_condition(problem.goal, fc)
        for atom in problem.initial:
            if isinstance(atom, pddl.Atom):
                error = check_atom(atom) or error
            else:
                error = check_function(atom.variable) or error
    error = error or pc.error or fc.error or ec.error

    return not error


class ExpressionArgumentChecker(pddl.RecursiveNumericExpressionVisitor):
    """
    Checks that the arguments of every function call is defined.
    """

    def __init__(self, arguments: dict[str, int]):
        self.arguments: dict[str, int] = arguments
        self.error: bool = False

    def visit_function_call(self, expr: pddl.FunctionCall) -> bool:
        for arg in expr.arguments:
            if arg.name not in self.arguments:
                _error(
                    sys.stderr,
                    f"Argument {arg.name} in function call {expr} is not defined.",
                )
                self.error = True
        return False

    def visit_atomic(self, expr: pddl.NumericExpression) -> bool:
        return False


class ConditionArgumentChecker(pddl.RecursiveBooleanExpressionVisitor):
    """
    Checks that the arguments of every atom and every function call
    appearing in some numeric condition is defined.
    """

    def __init__(self, arguments: dict[str, int]):
        self.arguments: dict[str, int] = arguments
        self.error: bool = False

    def visit_atom(self, atom: pddl.Atom) -> bool:
        for arg in atom.arguments:
            if arg.name not in self.arguments:
                _error(
                    sys.stderr, f"Argument {arg.name} in atom {atom} is not defined."
                )
                self.error = True
        return False

    def visit_numeric_condition(self, cond: pddl.NumericComparison) -> bool:
        ec = ExpressionArgumentChecker(self.arguments)
        cond.lhs.traverse(ec)
        cond.rhs.traverse(ec)
        self.error = self.error or ec.error
        return False

    def visit_atomic(self, x: pddl.BooleanExpression) -> bool:
        return False

    def visit_quantification(self, q: pddl.Quantification) -> bool:
        for x in q.parameters:
            self.arguments[x.name] = self.arguments.get(x.name, 0) + 1
        super().visit_quantification(q)
        for x in q.parameters:
            self.arguments[x.name] -= 1
            if self.arguments[x.name] == 0:
                del self.arguments[x.name]
        return False


class EffectArgumentChecker(pddl.RecursiveActionEffectVisitor):
    """
    Checks that all atoms and function calls have defined arguments.
    """

    def __init__(self, arguments: dict[str, int]):
        self.arguments: dict[str, int] = arguments
        self.error: bool = False

    def visit_atom_effect(self, atom: pddl.AtomEffect) -> bool:
        for arg in atom.arguments:
            if arg.name not in self.arguments:
                _error(
                    sys.stderr,
                    f"Argument {arg.name} of atom effect {atom} is not defined.",
                )
                self.error = True
        return False

    def visit_universal_effect(self, univ: pddl.UniversalEffect) -> bool:
        for arg in univ.parameters:
            self.arguments[arg.name] = self.arguments.get(arg.name, 0) + 1
        super(EffectArgumentChecker, self).visit_universal_effect(univ)
        for arg in univ.parameters:
            self.arguments[arg.name] -= 1
            if self.arguments[arg.name] == 0:
                del self.arguments[arg.name]
        return False

    def visit_conditional_effect(self, cond: pddl.ConditionalEffect) -> bool:
        cv = ConditionArgumentChecker(self.arguments)
        cond.condition.traverse(cv)
        self.error = self.error or cv.error
        return super(EffectArgumentChecker, self).visit_conditional_effect(cond)

    def visit_numeric_effect(self, eff: pddl.NumericEffect) -> bool:
        ev = ExpressionArgumentChecker(self.arguments)
        eff.variable.traverse(ev)
        eff.expression.traverse(ev)
        self.error = self.error or ev.error
        return False

    def visit_probabilistic_effect(self, eff: pddl.ProbabilisticEffect) -> bool:
        ev = ExpressionArgumentChecker(self.arguments)
        for outcome in eff.outcomes:
            outcome.probability.traverse(ev)
            self.error = self.error or ev.error
        return super(EffectArgumentChecker, self).visit_probabilistic_effect(eff)


def _argmap(args: tuple[pddl.ArgumentDefinition]) -> dict[str, int]:
    return {arg.name: 1 for arg in args}


def _insert_args(
    args: dict[str, int], new_args: tuple[pddl.ArgumentDefinition]
) -> None:
    for arg in new_args:
        args[arg.name] = args.get(arg.name, 0) + 1


def _delete_args(
    args: dict[str, int], new_args: tuple[pddl.ArgumentDefinition]
) -> None:
    for arg in new_args:
        args[arg.name] -= 1
        if args[arg.name] == 0:
            del args[arg.name]


def check_domain_variable_constant_references(
    domain: pddl.Domain, file=sys.stderr
) -> bool:
    """
    Check that all atoms and function calls in the domain reference defined arguments
    (constants and variables).
    """
    args = _argmap(domain.constants)
    cc = ConditionArgumentChecker(args)
    ec = EffectArgumentChecker(args)
    for action in domain.actions:
        _insert_args(args, action.parameters)
        action.precondition.traverse(cc)
        action.effect.traverse(ec)
        _delete_args(args, action.parameters)
    for predicate in domain.derived_predicates:
        _insert_args(args, predicate.predicate.parameters)
        predicate.condition.traverse(cc)
        _delete_args(args, predicate.predicate.parameters)
    return not cc.error and not ec.error


def check_problem_variable_object_references(
    domain: pddl.Domain, problem: pddl.Problem, file=sys.stderr
) -> bool:
    """
    Checks that all atoms and function calls in the problem reference existing
    objects or variables.
    """
    args = _argmap(domain.constants)
    _insert_args(args, problem.objects)
    cc = ConditionArgumentChecker(args)
    nc = ExpressionArgumentChecker(args)
    problem.goal.traverse(cc)
    if problem.metric:
        problem.metric.expression.traverse(nc)
    return not cc.error and not nc.error


def make_checks(domain: pddl.Domain, problem: pddl.Problem, file=sys.stderr) -> bool:
    """
    Performs syntactic sanity checks (type, function, predicate, variable, object, and
    action names and references).
    """
    return all([
        unique_predicates(domain, file=file),
        unique_functions(domain, file=file),
        unique_action_names(domain, file=file),
        unique_type_names(domain, file=file),
        unique_object_names(domain.constants, file=file),
        unique_object_names(problem.objects, file=file),
        unique_variable_names(domain, problem, file=file),
        no_reserved_functions(domain, file=file),
        type_hierarchy(domain, file=file),
        predicate_references(domain, problem, file=file),
        check_domain_variable_constant_references(domain, file),
        check_problem_variable_object_references(domain, problem, file),
    ])



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, NamedTuple

from plado.datalog.numeric import NumericConstraint

RelationPositionRef = NamedTuple(
    "RelationPositionRef", [("relation", int), ("position", int)]
)

ArgRefMap = dict[int, set[RelationPositionRef]]


RelationArgs = tuple[int]


def merge_arg_maps(args0: ArgRefMap, args1: ArgRefMap) -> ArgRefMap:
    result = {k: set(refs) for (k, refs) in args0.items()}
    for k, refs in args1.items():
        if k not in result:
            result[k] = set()
        result[k] |= refs
    return result


def get_arg_map(relation_id: int, relation_args: tuple[int]) -> ArgRefMap:
    return {
        int(x): set([RelationPositionRef(relation_id, pos)])
        for (pos, x) in enumerate(relation_args)
    }


def is_contained(args: Iterable[int], arg_map: ArgRefMap) -> bool:
    return all((x in arg_map for x in args))


class QNode(ABC):
    def __init__(self, cost: int):
        self.cost: int = cost

    @abstractmethod
    def lheight(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def get_relations(self) -> set[int]:
        raise NotImplementedError()

    @abstractmethod
    def get_registers(self) -> set[int]:
        raise NotImplementedError()

    @abstractmethod
    def get_argument_map(self) -> ArgRefMap:
        raise NotImplementedError()

    @abstractmethod
    def get_jg_neighbors(self) -> set[int]:
        raise NotImplementedError()

    @abstractmethod
    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        raise NotImplementedError()


class EmptyTupleNode(QNode):
    def __init__(self):
        super().__init__(0)

    def __str__(self):
        return "{()}"

    def lheight(self) -> int:
        return 1

    def get_relations(self) -> set[int]:
        return set()

    def get_registers(self) -> set[int]:
        return set()

    def get_argument_map(self) -> ArgRefMap:
        return ArgRefMap()

    def get_jg_neighbors(self) -> set[int]:
        return set()

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_empty_tuple(self)


class RegisterNode(QNode):
    def __init__(
        self,
        register_id: int,
        relation_args: tuple[int],
    ):
        super().__init__(0)
        self.register_id: int = register_id
        self.relation_args: tuple[int] = relation_args

    def lheight(self) -> int:
        return 1

    def get_relations(self) -> set[int]:
        return set()

    def get_registers(self) -> set[int]:
        return set([self.register_id])

    def get_argument_map(self) -> ArgRefMap:
        return get_arg_map(None, self.relation_args)

    def get_jg_neighbors(self) -> set[int]:
        return set()

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_register(self)

    def __str__(self) -> str:
        return (
            f"(R{self.register_id} {' '.join((f'?x{i}' for i in self.relation_args))})"
        )


class LeafNode(QNode):
    def __init__(
        self,
        relation_id: int,
        relation_args: tuple[int],
        jg_arcs: Iterable[int],
        cost: int,
    ):
        super().__init__(cost)
        self.relation_id: int = relation_id
        self.relation_args: RelationArgs = RelationArgs((int(x) for x in relation_args))
        self.jg_arcs: set[int] = set(jg_arcs)

    def lheight(self) -> int:
        return 1

    def __str__(self) -> str:
        return (
            f"(P{self.relation_id} {' '.join((f'?x{i}' for i in self.relation_args))})"
        )

    def get_relations(self) -> set[int]:
        return set([self.relation_id])

    def get_registers(self) -> set[int]:
        return set()

    def get_jg_neighbors(self) -> set[int]:
        return self.jg_arcs

    def get_argument_map(self) -> ArgRefMap:
        return get_arg_map(self.relation_id, self.relation_args)

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_leaf(self)


class InnerNode(QNode):
    def __init__(self, left: QNode, right: QNode, cost: int):
        super().__init__(cost)
        self.relations: set[int] = left.get_relations() | right.get_relations()
        self.jg_arcs: set[int] = left.get_jg_neighbors() | right.get_jg_neighbors()
        self.arg_map: ArgRefMap = merge_arg_maps(
            left.get_argument_map(), right.get_argument_map()
        )
        self.left: QNode = left
        self.right: QNode = right
        self.shared_args: set[int] = set(
            left.get_argument_map().keys() & right.get_argument_map().keys()
        )

    def lheight(self) -> int:
        return self.left.lheight() + 1

    def get_relations(self) -> set[int]:
        return self.relations

    def get_registers(self) -> set[int]:
        return self.left.get_registers() | self.right.get_registers()

    def get_jg_neighbors(self) -> set[int]:
        return self.jg_arcs

    def get_argument_map(self) -> ArgRefMap:
        return self.arg_map


class JoinNode(InnerNode):

    def __str__(self) -> str:
        return f"[{str(self.left)}] . [{str(self.right)}]"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_join(self)


class DifferenceNode(JoinNode):

    def __str__(self) -> str:
        return f"[{str(self.left)}] - [{str(self.right)}]"

    def lheight(self) -> int:
        return self.left.lheight()

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_difference(self)


class ProductNode(InnerNode):

    def __str__(self) -> str:
        return f"[{str(self.left)}] X [{str(self.right)}]"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_product(self)


class WrappingNode(QNode):
    def __init__(self, child: QNode):
        super().__init__(child.cost)
        self.child: QNode = child

    def lheight(self) -> int:
        return self.child.lheight()

    def get_registers(self) -> set[int]:
        return self.child.get_registers()

    def get_relations(self) -> set[int]:
        return self.child.get_relations()

    def get_jg_neighbors(self) -> set[int]:
        return self.child.get_jg_neighbors()

    def get_argument_map(self) -> ArgRefMap:
        return self.child.get_argument_map()


class PredicateNode(WrappingNode):
    def __init__(self, child: QNode, variable_id: int, value_ref: int):
        super().__init__(child)
        self.variable_id = variable_id
        self.value_ref = value_ref


class EqualityPredicateNode(PredicateNode):

    def __str__(self) -> str:
        return f"[{str(self.child)}] where ?x{self.variable_id}==?x{self.value_ref}"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_equality(self)


class InequalityPredicateNode(PredicateNode):

    def __str__(self) -> str:
        return f"[{str(self.child)}] where ?x{self.variable_id}!=?x{self.value_ref}"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_inequality(self)


class SelectPredicateNode(PredicateNode):

    def __str__(self) -> str:
        return f"[{str(self.child)}] where ?x{self.variable_id}=={self.value_ref}"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_select(self)


class SelectNotPredicateNode(PredicateNode):

    def __str__(self) -> str:
        return f"[{str(self.child)}] where ?x{self.variable_id}!={self.value_ref}"

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_select_not(self)


class NumericConditionNode(WrappingNode):
    def __init__(self, child: QNode, constraint: NumericConstraint):
        super().__init__(child)
        self.constraint: NumericConstraint = constraint

    def __str__(self) -> str:
        return (f"[{str(self.child)} where {str(self.constraint)}]")

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_numeric(self)


class ProjectionNode(WrappingNode):
    def __init__(self, child: QNode, args: Iterable[int]):
        super().__init__(child)
        self.projection: tuple[int] = tuple((int(x) for x in args))
        assert all((x in child.get_argument_map() for x in self.projection))
        self.arg_map: ArgRefMap = {
            x: child.get_argument_map()[x] for x in self.projection
        }

    def accept(self, visitor: "QueryTreeVisitor") -> Any:
        return visitor.visit_projection(self)

    def get_argument_map(self) -> ArgRefMap:
        return self.arg_map

    def __str__(self) -> str:
        return (
            f"[{str(self.child)}] project onto"
            f" {', '.join((f'?x{i}' for i in self.projection))}"
        )


class QueryTreeVisitor(ABC):
    @abstractmethod
    def visit_generic(self, node: QNode) -> Any:
        raise NotImplementedError()

    def visit_register(self, node: LeafNode) -> Any:
        return self.visit_generic(node)

    def visit_leaf(self, node: LeafNode) -> Any:
        return self.visit_generic(node)

    def visit_binary_op(self, node: InnerNode) -> Any:
        return self.visit_generic(node)

    def visit_join(self, node: JoinNode) -> Any:
        return self.visit_binary_op(node)

    def visit_difference(self, node: DifferenceNode) -> Any:
        return self.visit_binary_op(node)

    def visit_product(self, node: ProductNode) -> Any:
        return self.visit_binary_op(node)

    def visit_predicate(self, node: PredicateNode) -> Any:
        return self.visit_generic(node)

    def visit_equality(self, node: EqualityPredicateNode) -> Any:
        return self.visit_predicate(node)

    def visit_inequality(self, node: InequalityPredicateNode) -> Any:
        return self.visit_predicate(node)

    def visit_select(self, node: SelectPredicateNode) -> Any:
        return self.visit_predicate(node)

    def visit_select_not(self, node: SelectNotPredicateNode) -> Any:
        return self.visit_predicate(node)

    def visit_projection(self, node: ProjectionNode) -> Any:
        return self.visit_generic(node)

    def visit_numeric(self, node: NumericConditionNode) -> Any:
        return self.visit_generic(node)

    def visit_empty_tuple(self, node: EmptyTupleNode) -> Any:
        return self.visit_generic(node)


class QueryTreeTransformer(QueryTreeVisitor):

    def visit_generic(self, node: QNode) -> QNode:
        return node

    def visit_binary_op(self, node: InnerNode) -> QNode:
        return node.__class__(
            node.left.accept(self), node.right.accept(self), node.cost
        )

    def visit_join(self, node: JoinNode) -> QNode:
        return self.visit_binary_op(node)

    def visit_difference(self, node: DifferenceNode) -> QNode:
        return self.visit_binary_op(node)

    def visit_product(self, node: ProductNode) -> QNode:
        return self.visit_binary_op(node)

    def visit_predicate(self, node: PredicateNode) -> QNode:
        return node.__class__(node.child.accept(self), node.variable_id, node.value_ref)

    def visit_equality(self, node: EqualityPredicateNode) -> QNode:
        return self.visit_predicate(node)

    def visit_inequality(self, node: InequalityPredicateNode) -> QNode:
        return self.visit_predicate(node)

    def visit_select(self, node: SelectPredicateNode) -> QNode:
        return self.visit_predicate(node)

    def visit_select_not(self, node: SelectNotPredicateNode) -> QNode:
        return self.visit_predicate(node)

    def visit_projection(self, node: ProjectionNode) -> QNode:
        return node.__class__(node.child.accept(self), node.arg_map.keys())

    def visit_numeric(self, node: NumericConditionNode) -> QNode:
        return node.__class__(node.child.accept(self), node.constraint)



#!/usr/bin/env python

import argparse

from plado.parser import parse_and_normalize
from plado.semantics.applicable_actions_generator import ApplicableActionsGenerator
from plado.semantics.goal_checker import GoalChecker
from plado.semantics.successor_generator import SuccessorGenerator
from plado.semantics.task import Task


def parse_plan_file(task: Task, path: str) -> list[tuple[int, tuple[int]]]:
    result = []
    with open(path, encoding="ascii") as f:
        for l in f:
            i = l.find(";")
            if i >= 0:
                l = l[:i]
            l = l.strip()
            if len(l) > 0:
                assert l[0] == "(" and l[-1] == ")"
                seg = l[1:-1].split(" ")
                action = [i for i, a in enumerate(task.actions) if a.name == seg[0]]
                args = [task.objects.index(seg[i]) for i in range(1, len(seg))]
                result.append((action[0], tuple(args)))
    return result


def main():
    p = argparse.ArgumentParser()
    p.add_argument("domain")
    p.add_argument("problem")
    p.add_argument("plan")
    p.add_argument("--verbose", "-v", action="store_true")
    p.add_argument("--very-verbose", "-vv", action="store_true")

    args = p.parse_args()

    args.verbose = args.verbose or args.very_verbose

    domain, problem = parse_and_normalize(args.domain, args.problem)
    task = Task(domain, problem)
    plan = parse_plan_file(task, args.plan)

    sgen = SuccessorGenerator(task)
    aops_gen = ApplicableActionsGenerator(task)
    goal_check = GoalChecker(task)

    state = task.initial_state
    if args.verbose:
        print("state0:", task.dump_state(state))
    for i, (a, params) in enumerate(plan):
        if args.verbose:
            print(f"action{i}:", task.dump_action(a, params))
        aops = tuple(aops_gen(state))
        if (a, params) not in aops:
            print(
                f"Error at step {i}: action",
                task.dump_action(a, params),
                "is not applicable in state ",
                task.dump_state(state),
            )
            if args.very_verbose:
                print(f"State has {len(aops)} applicable actions")
                for b, bargs in aops:
                    print(task.dump_action(b, bargs))
            return 1
        successors = sgen(state, (a, params))
        assert len(successors) == 1, "Cannot validate plan in probabilistic domain"
        state = successors[0][0]
        if args.verbose:
            print(f"state{i+1}:", task.dump_state(state))

    if not goal_check(state):
        print("Error not a goal state:", task.dump_state(state))
        return 1

    print("Plan is valid.")

    return 0


if __name__ == "__main__":
    main()



from collections.abc import Callable, Iterable

from plado.datalog.evaluator.join_graph import JoinGraph
from plado.datalog.evaluator.query_tree import (
    ArgRefMap,
    DifferenceNode,
    EmptyTupleNode,
    JoinNode,
    LeafNode,
    ProductNode,
    QNode,
    is_contained,
)
from plado.utils.graph import tarjan
from plado.utils.union_find import UnionFind

Nodes = dict[int, QNode]


class GreedyOptimizer:
    def __init__(
        self,
        cost_estimator: Callable[
            # relations, relations' args, new relation, arg map, join args
            [set[int], ArgRefMap, set[int], ArgRefMap],
            # cost
            int,
        ],
    ):
        self.jg: JoinGraph | None = None
        self.cost_estimator = cost_estimator

    def _select(
        self, nodes: Nodes, get_candidates: Callable[[QNode], Iterable[int]]
    ) -> tuple[int, int, int]:
        """
        Greedily select two nodes in the current list to join. Evaluate cost
        function for all pairs, and return the cheapest option.
        """
        assert len(nodes) > 1
        best_cost, left, right = None, None, None
        for node_id, node in nodes.items():
            # consider only nodes with intersecting argument list (i.e., doing a
            # join rather than the cross product of the two associated relations)
            candidates = get_candidates(node)
            # evaluate cost function for each candidate
            for candidate in candidates:
                if candidate == node_id:
                    continue
                assert candidate in nodes
                cand_node = nodes[candidate]
                cost = self.cost_estimator(
                    node.get_relations(),
                    node.get_argument_map(),
                    cand_node.get_relations(),
                    cand_node.get_argument_map(),
                )
                # update incumbent solution as necessary
                if best_cost is None or cost < best_cost:
                    best_cost, left, right = cost, node_id, candidate
        return best_cost, left, right

    def _select_for_join(
        self, nodes: Nodes, jg_to_jt: UnionFind
    ) -> tuple[int, int, int]:
        """
        Greedily try out all possible combinations of two nodes in the current
        list with interesecting arguments (hence join). Evaluate cost function
        for all pairs, and return the cheapest option.
        """

        def get_successors(node: QNode) -> Iterable[int]:
            # consider only nodes with intersecting argument list (i.e., doing a
            # join rather than the cross product of the two associated relations)
            candidates = set((jg_to_jt[x] for x in node.jg_arcs))
            for candidate in candidates:
                # candidate may not be in nodes if referenced node is negative
                if candidate not in nodes:
                    assert self.jg.nodes[candidate].negated
                else:
                    yield candidate

        return self._select(nodes, get_successors)

    def _select_for_product(self, nodes: Nodes) -> tuple[int, int, int]:
        """
        Similar to _select_for_join; but here try out *all* possible combinations of
        nodes. This is necessary, if the sub-graph spanned by nodes is not
        connected. The considered join sub-graph might get disconnected as part
        of removing negative nodes.
        """

        def get_successors(node: QNode) -> Iterable[int]:
            niter = nodes.items()
            # prevent duplicate combinations (combine node only with proceeding
            # nodes in the node list)
            for node_id, other_node in niter:
                if other_node is node:
                    break
            for node_id, _ in niter:
                yield node_id

        return self._select(nodes, get_successors)

    def _join_nodes(
        self,
        cost: int,
        left: QNode,
        right: QNode,
        subtraction: dict[tuple[int], list[LeafNode]],
        join_class: type[QNode],
    ) -> QNode:
        """
        Join the two nodes and apply any subtraction whose arguments are contained
        in the join result.
        """
        assert left is not right
        if left.cost > right.cost:
            left, right = right, left
        jnode = join_class(left, right, cost)
        return self._apply_subtractions(jnode, subtraction)

    def _apply_subtractions(
        self, node: QNode, subtraction: dict[tuple[int], list[LeafNode]]
    ) -> QNode:
        """
        Immediately apply all ``set minus joins'' (subtractions) for the
        relations in the subtraction list whose parameters are fully contained
        in node.
        """
        for args in list(subtraction.keys()):
            if is_contained(args, node.get_argument_map()):
                for leaf in subtraction[args]:
                    node = DifferenceNode(
                        node,
                        leaf,
                        max(node.cost, leaf.cost),
                    )
                del subtraction[args]
        return node

    def _select_and_join(
        self,
        nodes: Nodes,
        subtraction: dict[tuple[int], list[LeafNode]],
        jg_to_jt: UnionFind,
    ) -> None:
        """
        Select two nodes from the node list and replace them by their join.
        """
        cost, left, right = self._select_for_join(nodes, jg_to_jt)
        join_class = JoinNode
        # if join no longer possible do cross product
        if cost is None:
            cost, left, right = self._select_for_product(nodes)
            assert cost is not None
            join_class = ProductNode
        left_node = nodes[left]
        right_node = nodes[right]
        del nodes[left]
        del nodes[right]
        new_id = jg_to_jt.merge(left, right)
        nodes[new_id] = self._join_nodes(
            cost, left_node, right_node, subtraction, join_class
        )

    def _create_jt_nodes(
        self, node_ids: list[int]
    ) -> tuple[dict[int, QNode], dict[tuple[int], list[LeafNode]]]:
        """
        Create the QNodes for the join-graph nodes whose ids are contained in
        node_ids. Negative nodes are handled and returned separately. This function
        returns a pair, whose first element is the list of all QNodes for the
        referenced non-negative join nodes. The second element is a mapping from
        the relations' arguments to a list of QNodes which must be subtracted
        from the joins.
        """
        assert len(node_ids) > 0
        # create leaf nodes (relation lookups)
        leafs: list[QNode] = [
            LeafNode(
                self.jg.nodes[node_id].relation_id,
                self.jg.nodes[node_id].args,
                (succ.node_id for succ in self.jg.arcs[node_id]),
                0,
            )
            for node_id in node_ids
        ]
        for leaf in leafs:
            leaf.cost = self.cost_estimator(
                leaf.get_relations(), leaf.get_argument_map(), set([]), {}
            )
        # seperate out negative nodes and store in ``subtraction'' mapping (for
        # easier implementation of the apply function)
        neg_nodes: list[LeafNode] = [
            leafs[i] for i, x in enumerate(node_ids) if self.jg.nodes[x].negated
        ]
        # mapping from argument list to negative nodes
        subtraction = {}
        for node in neg_nodes:
            sargs = tuple(sorted(node.get_argument_map().keys()))
            if sargs not in subtraction:
                subtraction[sargs] = []
            subtraction[sargs].append(node)
        # construct node list, immediately apply subtractions if possible
        nodes: Nodes = {
            x: self._apply_subtractions(leafs[i], subtraction)
            for i, x in enumerate(node_ids)
            if not self.jg.nodes[x].negated
        }
        return nodes, subtraction

    def _compute_join_ordering(self, node_ids: list[int]) -> QNode:
        """
        Serialize the given join-graph nodes into a tree.
        """
        assert len(node_ids) > 0
        # create leaf nodes
        nodes, subtraction = self._create_jt_nodes(node_ids)
        if len(nodes) == 0:
            assert len(subtraction) == 1 and tuple() in subtraction
            return self._apply_subtractions(EmptyTupleNode(), subtraction)
        # mapping storing the relation of the join-graph node ids jg_to_jt
        # ids of the nodes in the join tree
        jg_to_jt = UnionFind(len(self.jg.nodes))
        # merge two nodes until only one is left
        while len(nodes) > 1:
            self._select_and_join(nodes, subtraction, jg_to_jt)
        assert len(nodes) == 1
        return tuple(nodes.values())[0]

    def _compute_components(self, jg: JoinGraph) -> list[list[int]]:
        """
        Decompose join graph into connected components. Each connected component
        is joined individually. Performing the cross product of all joins
        afterwards.
        """
        components: list[list[int]] = []
        group: list[int] = []
        visited: list[bool] = [False for i in range(len(jg.arcs))]

        def get_successors(node: int) -> Iterable[int]:
            assert not visited[node]
            visited[node] = True
            group.append(node)
            return (
                succ_node.node_id
                for succ_node in jg.arcs[node]
                if not visited[succ_node.node_id]
            )

        for node in range(len(jg.arcs)):
            if visited[node]:
                continue
            tarjan(node, get_successors)
            components.append(list(group))
            group.clear()
        return components

    def __call__(self, join_graph: JoinGraph) -> QNode:
        """
        Serialize the given join graph into a tree.
        """
        self.jg = join_graph
        components: list[list[int]] = self._compute_components(self.jg)
        assert len(components) > 0
        nodes: list[QNode] = [
            self._compute_join_ordering(component) for component in components
        ]
        node = nodes[0]
        for i in range(1, len(nodes)):
            node = ProductNode(node, nodes[i], node.cost * nodes[i].cost)
        return node



from typing import TypeVar

from plado import pddl
from plado.parser.tokenizer import Category, Token

T = TypeVar("T")


class ParseError(ValueError):
    def __init__(self, message):
        super().__init__(message)


class UnexpectedEndOfFile(ParseError):
    def __init__(self, message: str = ""):
        super().__init__("Unexpected end of file. " + message)


class UnexpectedToken(ParseError):
    def __init__(self, token: Token, expected: str = ""):
        super().__init__(f"Received unexpected token {token}. {expected}")


class LookaheadStreamer:
    def __init__(self, token_stream):
        self.lookahead_buffer = []
        self.stream = token_stream

    def __iter__(self):
        """
        Returns self reference.
        """
        return self

    def __next__(self) -> Token:
        """
        Returns and consumes next token from the stream.
        """
        if len(self.lookahead_buffer) > 0:
            tok, self.lookahead_buffer = (
                self.lookahead_buffer[0],
                self.lookahead_buffer[1:],
            )
            return tok
        return next(self.stream)

    def __getitem__(self, i: int) -> Token:
        """
        @see at()
        """
        return self.at(i)

    def at(self, i: int) -> Token:
        """
        Lookahead the i-th next token (i=0 is the next token, and so on).
        Does change the external state of the iterator, i.e., calling
        at(i) multiple times will return the exact same token.
        """
        assert i >= 0
        try:
            while i >= len(self.lookahead_buffer):
                self.lookahead_buffer.append(next(self.stream))
        except StopIteration as exc:
            raise UnexpectedEndOfFile() from exc
        return self.lookahead_buffer[i]

    def shift(self, n: int) -> list[Token]:
        """
        Returns and consumes the next n tokens from the stream.
        """
        assert n >= 0
        if n == 0:
            return []
        # fetch tokens
        self.at(n - 1)
        result, self.lookahead_buffer = (
            self.lookahead_buffer[:n],
            self.lookahead_buffer[n:],
        )
        return result

    def lookahead(self, n: int) -> list[Token]:
        """
        Returns the next n tokens from the stream without consuming them.
        """
        assert n >= 0
        if n == 0:
            return []
        self.at(n - 1)
        return self.lookahead_buffer[:n]


def parse_requirements(tokens: LookaheadStreamer) -> list[str]:
    tok = next(tokens)
    requirements = []
    while tok.cat != Category.RBRACK:
        if tok.cat != Category.REQUIREMENT_CONSTANT:
            raise UnexpectedToken(tok, "Expected requirement specification")
        requirements.append(tok.tok.lower())
        tok = next(tokens)
    return requirements


def _parse_typed_list(
    tokens: LookaheadStreamer,
    category: Category,
    object_class: type[T],
    hint: str = None,
) -> list[T]:
    tok = next(tokens)
    var_list = []
    name_buffer = []
    while tok.cat != Category.RBRACK:
        if tok.cat == Category.MINUS:
            tok = next(tokens)
            if tok.cat != Category.NAME:
                raise UnexpectedToken(tok, "Expected type name.")
            var_list.extend((
                object_class(o.name, tok.tok)
                for o in name_buffer
            ))
            name_buffer = []
        elif tok.cat == category:
            name_buffer.append(object_class(tok.tok, "object"))
        else:
            raise UnexpectedToken(
                tok, "Expected name." if hint is None else f"Expected {hint} name."
            )
        tok = next(tokens)
    var_list.extend(name_buffer)
    return var_list


def parse_constants(tokens: LookaheadStreamer) -> list[pddl.ArgumentDefinition]:
    return _parse_typed_list(tokens, Category.NAME, pddl.ArgumentDefinition, "object")


def parse_types(tokens: LookaheadStreamer) -> list[pddl.Type]:
    return _parse_typed_list(tokens, Category.NAME, pddl.Type, "type")


def _parse_function(tokens: LookaheadStreamer, function_class: type[T], hint: str) -> T:
    tok = next(tokens)
    if tok.cat != Category.NAME:
        raise UnexpectedToken(tok, f"Expected {hint} name.")
    name = tok.tok
    tok = tokens.at(0)
    if tok.cat == Category.RBRACK:
        next(tokens)
        return function_class(name, [])
    parameters = _parse_typed_list(
        tokens, Category.VARIABLE, pddl.ArgumentDefinition, "variable"
    )
    return function_class(name, parameters)


def parse_predicates(tokens: LookaheadStreamer) -> list[pddl.Predicate]:
    tok = next(tokens)
    functions = []
    while tok.cat != Category.RBRACK:
        functions.append(_parse_function(tokens, pddl.Predicate, "predicate"))
        tok = next(tokens)
    return functions


def parse_functions(tokens: LookaheadStreamer) -> list[pddl.Function]:
    tok = next(tokens)
    functions = []
    while tok.cat != Category.RBRACK:
        functions.append(_parse_function(tokens, pddl.Function, "function"))
        tok = next(tokens)
        if tok.cat == Category.MINUS:
            tok = next(tokens)
            if tok.cat != Category.NAME or tok.tok.lower() != "number":
                raise UnexpectedToken(tok, "Expected 'number'")
            tok = next(tokens)
    return functions


def _parse_open(tokens: LookaheadStreamer) -> None:
    tok = next(tokens)
    if tok.cat != Category.LBRACK:
        raise UnexpectedToken(tok, "Expected '('")


def _parse_close(tokens: LookaheadStreamer) -> None:
    tok = next(tokens)
    if tok.cat != Category.RBRACK:
        raise UnexpectedToken(tok, "Expected closing ')'")


def _parse_term(token: Token) -> pddl.Argument:
    if token.cat == Category.NAME:
        return pddl.ObjectArgument(token.tok)
    if token.cat == Category.VARIABLE:
        return pddl.VariableArgument(token.tok)
    raise UnexpectedToken(token, "Expected object or variable.")


def _parse_atom(name: str, tokens: LookaheadStreamer, atom_class: type[T]) -> T:
    arguments = []
    tok = next(tokens)
    while tok.cat != Category.RBRACK:
        arguments.append(_parse_term(tok))
        tok = next(tokens)
    return atom_class(name, arguments)


def _parse_expression(tokens: LookaheadStreamer) -> pddl.NumericExpression:
    tok = next(tokens)
    if tok.cat == Category.CONSTANT:
        return pddl.NumericConstant(tok.tok)
    if tok.cat != Category.LBRACK:
        raise UnexpectedToken("Expected '('")
    tok = next(tokens)
    if tok.cat == Category.NAME:
        return _parse_atom(tok.tok, tokens, pddl.FunctionCall)
    if tok.cat == Category.MINUS:
        lhs = _parse_expression(tokens)
        if tok.cat == Category.RBRACK:
            next(tokens)
            return pddl.UnaryNegation(lhs)
        rhs = _parse_expression(tokens)
        _parse_close(tokens)
        return pddl.Subtraction(lhs, rhs)
    if tok.cat in [Category.PLUS, Category.TIMES, Category.DIVIDE]:
        lhs = _parse_expression(tokens)
        rhs = _parse_expression(tokens)
        _parse_close(tokens)
        if tok.cat == Category.PLUS:
            return pddl.Sum(lhs, rhs)
        if tok.cat == Category.TIMES:
            return pddl.Product(lhs, rhs)
        return pddl.Division(lhs, rhs)
    raise UnexpectedToken(tok, "Could not parse expression.")


def _parse_formula(tokens: LookaheadStreamer) -> pddl.BooleanExpression:
    tok = next(tokens)
    if tok.cat == Category.NAME:
        return _parse_atom(tok.tok, tokens, pddl.Atom)
    if tok.cat == Category.EQUAL:
        tok = tokens[0]
        try:
            lhs = _parse_term(tok)
            next(tokens)
            rhs = _parse_term(next(tokens))
            _parse_close(tokens)
            return pddl.Atom("=", [lhs, rhs])
        except ValueError:
            lhs = _parse_expression(tokens)
            rhs = _parse_expression(tokens)
            _parse_close(tokens)
            return pddl.Equals(lhs, rhs)
    if tok.cat == Category.NOT:
        _parse_open(tokens)
        sub_formula = _parse_formula(tokens)
        _parse_close(tokens)
        return pddl.Negation(sub_formula)
    if tok.cat in (Category.AND, Category.OR):
        sub_formulas = []
        while tokens[0].cat != Category.RBRACK:
            sub_formulas.append(_parse_condition(tokens))
        _parse_close(tokens)
        if tok.cat == Category.AND:
            return pddl.Conjunction(sub_formulas)
        return pddl.Disjunction(sub_formulas)
    if tok.cat == Category.IMPLY:
        lhs = _parse_condition(tokens)
        rhs = _parse_condition(tokens)
        _parse_close(tokens)
        return pddl.Disjunction([pddl.Negation(lhs), rhs])
    if tok.cat in (Category.EXISTS, Category.FORALL):
        _parse_open(tokens)
        parameters = _parse_typed_list(
            tokens, Category.VARIABLE, pddl.ArgumentDefinition, "variable"
        )
        sub_formula = _parse_condition(tokens)
        _parse_close(tokens)
        if tok.cat == Category.EXISTS:
            return pddl.Exists(parameters, sub_formula)
        return pddl.Forall(parameters, sub_formula)
    if tok.cat in [Category.LT, Category.LE, Category.GT, Category.GE]:
        lhs = _parse_expression(tokens)
        rhs = _parse_expression(tokens)
        _parse_close(tokens)
        if tok.cat == Category.LT:
            return pddl.Less(lhs, rhs)
        if tok.cat == Category.LE:
            return pddl.LessEqual(lhs, rhs)
        if tok.cat == Category.GT:
            return pddl.Greater(lhs, rhs)
        return pddl.GreaterEqual(lhs, rhs)
    raise UnexpectedToken(tok, "Could not parse formula.")


def _parse_condition(tokens: LookaheadStreamer) -> pddl.BooleanExpression:
    tok = next(tokens)
    if tok.cat != Category.LBRACK:
        raise UnexpectedToken(tok, "Expected '('")
    tok = tokens[0]
    if tok.cat == Category.RBRACK:
        next(tokens)
        return pddl.Truth()
    return _parse_formula(tokens)


def _parse_effect(tokens: LookaheadStreamer) -> pddl.ActionEffect:
    tok = next(tokens)
    if tok.cat != Category.LBRACK:
        raise UnexpectedToken(tok, "Expected '('")
    tok = next(tokens)
    if tok == Category.RBRACK:
        return pddl.ConjunctiveEffect([])
    if tok.cat == Category.NAME:
        return _parse_atom(tok.tok, tokens, pddl.AtomEffect)
    if tok.cat == Category.NOT:
        toks = tokens.shift(2)
        if toks[0].cat != Category.LBRACK:
            raise UnexpectedToken(tok, "Expected '('")
        if toks[1].cat != Category.NAME:
            raise UnexpectedToken(tok, "Expected predicate name")
        atom = _parse_atom(toks[1].tok, tokens, pddl.AtomEffect)
        _parse_close(tokens)
        return pddl.NegativeEffect(atom)
    if tok.cat in [
        Category.SCALE_UP,
        Category.SCALE_DOWN,
        Category.INCREASE,
        Category.DECREASE,
        Category.ASSIGN,
    ]:
        toks = tokens.shift(2)
        if toks[0].cat != Category.LBRACK:
            raise UnexpectedToken(tok, "Expected '('")
        if toks[1].cat != Category.NAME:
            raise UnexpectedToken(tok, "Expected function name")
        atom = _parse_atom(toks[1].tok, tokens, pddl.FunctionCall)
        expr = _parse_expression(tokens)
        _parse_close(tokens)
        if tok.cat == Category.ASSIGN:
            return pddl.NumericAssignEffect(atom, expr)
        if tok.cat == Category.SCALE_UP:
            return pddl.ScaleUpEffect(atom, expr)
        if tok.cat == Category.SCALE_DOWN:
            return pddl.ScaleDownEffect(atom, expr)
        if tok.cat == Category.INCREASE:
            return pddl.IncreaseEffect(atom, expr)
        if tok.cat == Category.DECREASE:
            return pddl.DecreaseEffect(atom, expr)
    elif tok.cat == Category.AND:
        effects = []
        while tokens[0].cat != Category.RBRACK:
            effects.append(_parse_effect(tokens))
        _parse_close(tokens)
        return pddl.ConjunctiveEffect(effects)
    elif tok.cat == Category.FORALL:
        _parse_open(tokens)
        parameters = _parse_typed_list(
            tokens, Category.VARIABLE, pddl.ArgumentDefinition, "variable"
        )
        effect = _parse_effect(tokens)
        _parse_close(tokens)
        return pddl.UniversalEffect(parameters, effect)
    elif tok.cat == Category.WHEN:
        cond = _parse_condition(tokens)
        effect = _parse_effect(tokens)
        _parse_close(tokens)
        return pddl.ConditionalEffect(cond, effect)
    elif tok.cat == Category.PROBABILISTIC:
        outcomes = []
        while tokens[0].cat != Category.RBRACK:
            tok = next(tokens)
            if tok.cat != Category.CONSTANT:
                raise UnexpectedToken(tok, "Expected probability constant.")
            prob = pddl.NumericConstant(tok.tok)
            effect = _parse_effect(tokens)
            outcomes.append(pddl.ProbabilisticOutcome(prob, effect))
        _parse_close(tokens)
        return pddl.ProbabilisticEffect(outcomes)
    raise UnexpectedToken(tok, "Could not parse action effect.")


def parse_action(tokens: LookaheadStreamer) -> pddl.Action:
    tok = next(tokens)
    if tok.cat != Category.NAME:
        raise UnexpectedToken(tok, "Expected action name")
    atok = tok
    tok = next(tokens)
    parameters = None
    precondition = None
    effect = None
    while tok.cat != Category.RBRACK:
        if tok.cat == Category.PARAMETERS:
            if parameters is not None:
                raise ParseError(f"Multiply defined :parameters at {tok}")
            tok = next(tokens)
            parameters = _parse_typed_list(
                tokens, Category.VARIABLE, pddl.ArgumentDefinition, "variable"
            )
        elif tok.cat == Category.PRECONDITION:
            if precondition is not None:
                raise ParseError(f"Multiply defined :precondition at {tok}")
            precondition = _parse_condition(tokens)
        elif tok.cat == Category.EFFECT:
            if effect is not None:
                raise ParseError(f"Multiply defined :effect at {tok}")
            effect = _parse_effect(tokens)
        else:
            raise UnexpectedToken(tok, "Could not parse action.")
        tok = next(tokens)
    if effect is None:
        raise ParseError(f"Error while parsing action at {tok}: no effect specified")
    if parameters is None:
        parameters = []
    return pddl.Action(atok.tok, parameters, precondition, effect)


def parse_derived_predicate(tokens: LookaheadStreamer) -> pddl.DerivedPredicate:
    _parse_open(tokens)
    predicate = _parse_function(tokens, pddl.Predicate, "predicate")
    condition = _parse_condition(tokens)
    _parse_close(tokens)
    return pddl.DerivedPredicate(predicate, condition)


def parse_domain(tokens: LookaheadStreamer) -> pddl.Domain:
    domain_header = tokens.shift(6)
    if domain_header[0].cat != Category.LBRACK:
        raise UnexpectedToken(domain_header[0], "Expected '('")
    if domain_header[1].cat != Category.DEFINE:
        raise UnexpectedToken(domain_header[1], "Expected define")
    if domain_header[2].cat != Category.LBRACK:
        raise UnexpectedToken(domain_header[2], "Expected (")
    if domain_header[3].cat != Category.DOMAIN_DEF:
        raise UnexpectedToken(domain_header[3], "Expected domain")
    if domain_header[4].cat != Category.NAME:
        raise UnexpectedToken(domain_header[4], "Expected a domain name")
    if domain_header[5].cat != Category.RBRACK:
        raise UnexpectedToken(domain_header[5], "Expected )")
    tok = next(tokens)
    domain_name = domain_header[4].tok
    requirements = None
    constants = None
    types = None
    functions = None
    predicates = None
    derived_predicates = []
    actions = []
    try:
        while tok.cat != Category.RBRACK:
            if tok.cat != Category.LBRACK:
                raise UnexpectedToken(tok, "Expected '('")
            tok = next(tokens)
            if tok.cat == Category.REQUIREMENTS:
                if requirements is not None:
                    raise ParseError("Multiply defined :requirements tags")
                requirements = parse_requirements(tokens)
            elif tok.cat == Category.CONSTANTS:
                if constants is not None:
                    raise ParseError("Multiply defined :constants tags")
                constants = parse_constants(tokens)
            elif tok.cat == Category.FUNCTIONS:
                if functions is not None:
                    raise ParseError("Multiply defined :functions tags")
                functions = parse_functions(tokens)
            elif tok.cat == Category.TYPES:
                if types is not None:
                    raise ParseError("Multiply defined :types tags")
                types = parse_types(tokens)
            elif tok.cat == Category.PREDICATES:
                if predicates is not None:
                    raise ParseError("Multiply defined :predicates tags")
                predicates = parse_predicates(tokens)
            elif tok.cat == Category.DERIVED:
                derived_predicates.append(parse_derived_predicate(tokens))
            elif tok.cat == Category.ACTION:
                actions.append(parse_action(tokens))
            else:
                raise UnexpectedToken(tok, "Could not parse domain")
            tok = next(tokens)
    except StopIteration as exc:
        raise UnexpectedEndOfFile("Missing closing ')'") from exc
    if predicates is None or len(predicates) == 0:
        raise ParseError("No predicates specified")
    if len(actions) == 0:
        raise ParseError("No actions specified")
    # if types is None:
    #     types = [pddl.Type("object", None)]
    # if all((t.name != "object" for t in types)):
    #     types.append(pddl.Type("object", None))
    # predicates.append(pddl.Predicate("=", [pddl.Parameter("?x", "object"), pddl.Parameter("?y", "object")]))
    # functions = functions or []
    # if all((f.name != "reward" for f in functions)):
    #     functions.append(pddl.Function("reward", []))
    # if all((f.name != "total-cost" for f in functions)):
    #     functions.append(pddl.Function("total-cost", []))
    return pddl.Domain(
        domain_name,
        requirements or [],
        types or [],
        constants or [],
        predicates or [],
        functions or [],
        actions,
        derived_predicates,
    )


def parse_initial(
    tokens: LookaheadStreamer,
) -> list[pddl.Atom | pddl.NumericAssignEffect]:
    initial = []
    tok = next(tokens)
    while tok.cat != Category.RBRACK:
        if tok.cat != Category.LBRACK:
            raise UnexpectedToken(tok, "Expected '('")
        tok = next(tokens)
        if tok.cat == Category.EQUAL:
            _parse_open(tokens)
            tok = next(tokens)
            if tok.cat != Category.NAME:
                raise UnexpectedToken(tok, "Expected function name")
            atom = _parse_atom(tok.tok, tokens, pddl.FunctionCall)
            tok = next(tokens)
            if tok.cat != Category.CONSTANT:
                raise UnexpectedToken(tok, "Expected number.")
            _parse_close(tokens)
            initial.append(
                pddl.NumericAssignEffect(atom, pddl.NumericConstant(tok.tok))
            )
        elif tok.cat == Category.NAME:
            atom = _parse_atom(tok.tok, tokens, pddl.Atom)
            initial.append(atom)
        else:
            raise UnexpectedToken(tok, "Could not parse initial state.")
        tok = next(tokens)
    return initial


def parse_metric(tokens: LookaheadStreamer) -> pddl.Metric:
    tok = next(tokens)
    if tok.cat not in (Category.MINIMIZE, Category.MAXIMIZE):
        raise UnexpectedToken(tok, "Expected minimize or maximize")
    expr = _parse_expression(tokens)
    if tok.cat == Category.MINIMIZE:
        return pddl.Metric(pddl.Metric.MINIMIZE, expr)
    return pddl.Metric(pddl.Metric.MAXIMIZE, expr)


def parse_problem(tokens: LookaheadStreamer) -> pddl.Problem:
    header = tokens.shift(6)
    if header[0].cat != Category.LBRACK:
        raise UnexpectedToken(header[0], "Expected '('")
    if header[1].cat != Category.DEFINE:
        raise UnexpectedToken(header[1], "Expected define")
    if header[2].cat != Category.LBRACK:
        raise UnexpectedToken(header[2], "Expected (")
    if header[3].cat != Category.PROBLEM_DEF:
        raise UnexpectedToken(header[3], "Expected problem")
    if header[4].cat != Category.NAME:
        raise UnexpectedToken(header[4], "Expected a problem name")
    if header[5].cat != Category.RBRACK:
        raise UnexpectedToken(header[5], "Expected )")
    tok = next(tokens)
    problem_name = header[4].tok
    domain_name = None
    objects = None
    initial_state = None
    goal = None
    goal_reward = None
    metric = None
    try:
        while tok.cat != Category.RBRACK:
            if tok.cat != Category.LBRACK:
                raise UnexpectedToken(tok, "Expected '('")
            tok = next(tokens)
            if tok.cat == Category.OBJECTS:
                if objects is not None:
                    raise ParseError("Multiply defined :objects tags")
                objects = _parse_typed_list(
                    tokens, Category.NAME, pddl.ArgumentDefinition, "object"
                )
            elif tok.cat == Category.INIT:
                if initial_state is not None:
                    raise ParseError("Multiply defined :init tags")
                initial_state = parse_initial(tokens)
            elif tok.cat == Category.GOAL:
                if goal is not None:
                    raise ParseError("Multiply defined :goal tags")
                goal = _parse_condition(tokens)
                _parse_close(tokens)
            elif tok.cat == Category.GOAL_REWARD:
                if goal_reward is not None:
                    raise ParseError("Multiply defined :goal-reward tags")
                goal_reward = _parse_expression(tokens)
                _parse_close(tokens)
            elif tok.cat == Category.METRIC:
                if metric is not None:
                    raise ParseError("Multiply defined :metric tags")
                metric = parse_metric(tokens)
                _parse_close(tokens)
            elif tok.cat == Category.DOMAIN:
                if domain_name is not None:
                    raise ParseError("Multiply defined :domain tags")
                tok = next(tokens)
                if tok.cat != Category.NAME:
                    raise UnexpectedToken(tok, "Expected domain name")
                _parse_close(tokens)
                domain_name = tok.tok
            else:
                raise UnexpectedToken(tok, "Could not parse problem")
            tok = next(tokens)
    except StopIteration as exc:
        raise UnexpectedEndOfFile("Missing closing ')'") from exc
    if domain_name is None:
        raise ParseError("No domain name specified.")
    if goal is None:
        raise ParseError("No goal specified.")
    objects = objects or []
    initial_state = initial_state or []
    # initial_state.extend((pddl.Atom("=", [pddl.ObjectArgument(obj), pddl.ObjectArgument(obj)]) for obj in objects))
    # initial_state.append(pddl.NumericAssignEffect(pddl.FunctionCall("total-cost", []), pddl.NumericConstant("0")))
    # initial_state.append(pddl.NumericAssignEffect(pddl.FunctionCall("reward", []), pddl.NumericConstant("0")))
    return pddl.Problem(
        problem_name,
        domain_name,
        objects or [],
        initial_state or [],
        goal,
        goal_reward,
        metric,
    )



from collections.abc import Iterable

import plado.datalog.program as datalog
from plado.datalog.evaluator import DatalogEngine
from plado.semantics.task import AddEffect, AtomsStore, FluentsStore, State, Task

GroundActionRef = tuple[int, tuple[int]]


class Grounder:
    def __init__(self, task: Task):
        program = task.create_datalog_program()
        self.task: Task = task
        self.offset = program.num_relations()
        # preconditions
        for action in task.actions:
            rel_id = program.add_relation(action.parameters)
            clause = datalog.Clause(
                datalog.Atom(
                    rel_id,
                    (datalog.Constant(var, True) for var in range(action.parameters)),
                ),
                [],
                [],
                [],
            )
            action.precondition.to_datalog(len(task.predicates), clause)
            # remove negative preconditions and constraints
            clause.neg_body = tuple()
            clause.constraints = tuple()
            program.add_clause(clause)
        # add effects
        for action_id, action in enumerate(task.actions):
            for prob_eff in action.effect.effects:
                for _, outcome in prob_eff.outcomes:
                    for cond_eff in outcome:
                        if not isinstance(cond_eff.effect, AddEffect):
                            continue
                        add_effect: AddEffect = cond_eff.effect
                        clause = datalog.Clause(
                            add_effect.atom.to_datalog(),
                            [],
                            [],
                            [],
                        )
                        cond_eff.condition.to_datalog(len(task.predicates), clause)
                        # add precondition applicability fact
                        clause.pos_body = clause.pos_body + tuple([
                            datalog.Atom(
                                self.offset + action_id,
                                (
                                    datalog.Constant(var, True)
                                    for var in range(action.parameters)
                                ),
                            )
                        ])
                        # remove negative conditions
                        clause.neg_body = tuple()
                        clause.constraints = tuple()
                        program.add_clause(clause)
        # prepare for datalog query
        atoms, fluents = self.task.prepare_for_query(
            self.task.initial_state.atoms, self.task.initial_state.fluents
        )
        atoms.extend((set() for _ in range(len(self.task.actions))))
        # find datalog model
        engine = DatalogEngine(program, len(task.objects))
        self.model = engine(atoms, fluents)

    def get_action_instances(self, action_id: int) -> set[tuple[int]]:
        return self.model[self.offset + action_id]

    def get_predicate_instances(self, predicate_id: int) -> set[tuple[int]]:
        return self.model[predicate_id]

    def get_ground_actions(self) -> list[tuple[int, tuple[int]]]:
        result = []
        for action_id in range(len(self.task.actions)):
            result.extend(
                ((action_id, params) for params in self.get_action_instances(action_id))
            )
        return result

    def get_ground_facts(self) -> list[tuple[int, tuple[int]]]:
        result = []
        for predicate_id in range(len(self.task.predicates)):
            result.extend((
                (predicate_id, params)
                for params in self.get_predicate_instances(predicate_id)
            ))
        return result



import itertools
from collections.abc import Callable, Iterable
from fractions import Fraction
from typing import TypeVar

from plado import pddl
from plado.pddl_utils import (
    get_type_closure,
    transform_all_conditions,
    transform_all_conditions_in_effect,
    transform_all_conditions_with_vars,
    visit_all_conditions_in_effect,
    visit_all_expressions,
)


class FunctionsCollector(pddl.RecursiveNumericExpressionVisitor):
    def __init__(self):
        self.functions: set[str] = set()

    def visit_generic(self, expr: pddl.NumericExpression):
        raise ValueError()

    def visit_atomic(self, expr: pddl.NumericExpression):
        pass

    def visit_function_call(self, expr: pddl.FunctionCall):
        self.functions.add(expr.name)


class StandardizeVariableNamesCondition(pddl.BooleanExpressionTransformer):
    def __init__(self):
        self.var_map: dict[str, int] = {}
        self.num_vars = 0

    def visit_generic(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        return formula.substitute({
            pddl.VariableArgument(name): pddl.VariableArgument(f"?x{num}")
            for (name, num) in self.var_map.items()
        })

    def visit_quantification(self, formula: pddl.Quantification) -> pddl.Quantification:
        old_map = dict(self.var_map)
        new_arguments: list[pddl.ArgumentDefinition] = []
        for arg in formula.parameters:
            new_arguments.append(
                pddl.ArgumentDefinition(f"?x{self.num_vars}", arg.type_name)
            )
            self.var_map[arg.name] = self.num_vars
            self.num_vars += 1
        result = formula.__class__(new_arguments, formula.sub_formula.traverse(self))
        self.var_map = old_map
        return result


class StandardizeVariableNamesEffect(pddl.ActionEffectTransformer):
    def __init__(self):
        self.var_map: dict[str, int] = {}
        self.num_vars = 0

    def visit_generic(self, effect: pddl.ActionEffect) -> pddl.ActionEffect:
        return effect.substitute({
            pddl.VariableArgument(var_name): pddl.VariableArgument(f"?x{num}")
            for (var_name, num) in self.var_map.items()
        })

    def visit_conditional_effect(
        self, effect: pddl.ConditionalEffect
    ) -> pddl.ConditionalEffect:
        cv = StandardizeVariableNamesCondition()
        cv.var_map = self.var_map
        cv.num_vars = self.num_vars
        cond = effect.condition.traverse(cv)
        self.num_vars += cv.num_vars
        return pddl.ConditionalEffect(cond, effect.effect.traverse(self))

    def visit_universal_effect(
        self, effect: pddl.UniversalEffect
    ) -> pddl.UniversalEffect:
        old_map = dict(self.var_map)
        new_arguments: list[pddl.ArgumentDefinition] = []
        for arg in effect.parameters:
            new_arguments.append(
                pddl.ArgumentDefinition(f"?x{self.num_vars}", arg.type_name)
            )
            self.var_map[arg.name] = self.num_vars
            self.num_vars += 1
        result = pddl.UniversalEffect(new_arguments, effect.effect.traverse(self))
        self.var_map = old_map
        return result


def standardize_variable_names(domain: pddl.Domain, problem: pddl.Problem):
    for action in domain.actions:
        cv = StandardizeVariableNamesCondition()
        ev = StandardizeVariableNamesEffect()
        new_parameters = []
        for arg in action.parameters:
            cv.var_map[arg.name] = len(new_parameters)
            ev.var_map[arg.name] = len(new_parameters)
            new_parameters.append(
                pddl.ArgumentDefinition(f"?x{len(new_parameters)}", arg.type_name)
            )
        cv.num_vars = len(new_parameters)
        ev.num_vars = len(new_parameters)
        action.parameters = new_parameters
        action.precondition = action.precondition.traverse(cv)
        action.effect = action.effect.traverse(ev)
    for predicate in domain.derived_predicates:
        cv = StandardizeVariableNamesCondition()
        new_parameters = []
        for arg in predicate.parameters:
            cv.var_map[arg.name] = len(new_parameters)
            new_parameters.append(
                pddl.ArgumentDefinition(f"?x{len(new_parameters)}", arg.type_name)
            )
        cv.num_vars = len(new_parameters)
        predicate.parameters = new_parameters
        predicate.condition = predicate.condition.traverse(cv)
    cv = StandardizeVariableNamesCondition()
    problem.goal = problem.goal.traverse(cv)


class ReplaceForall(pddl.BooleanExpressionTransformer):
    def visit_forall(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        return pddl.Negation(
            pddl.Exists(
                formula.parameters, pddl.Negation(formula.sub_formula.traverse(self))
            )
        )


class PushNegation(pddl.BooleanExpressionTransformer):
    def __init__(self):
        self.negations = 0

    def _apply(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        return formula.negate() if self.negations % 2 else formula

    def visit_negation(self, formula: pddl.Negation) -> pddl.BooleanExpression:
        self.negations += 1
        result = formula.sub_formula.traverse(self)
        self.negations -= 1
        return result

    def visit_quantification(
        self, formula: pddl.Quantification
    ) -> pddl.BooleanExpression:
        # stop propagation at quantifiers
        n, self.negations = self.negations, 0
        sub = formula.sub_formula.traverse(self)
        newq = formula.__class__(formula.parameters, sub)
        self.negations = n
        return pddl.Negation(newq) if n % 2 else newq

    def visit_bool_connector(
        self, formula: pddl.BooleanConnector
    ) -> pddl.BooleanExpression:
        subs = [x.traverse(self) for x in formula.sub_formulas]
        if self.negations % 2:
            return formula.inverse_connector()(subs)
        return formula.__class__(subs)

    def visit_truth(
        self, formula: pddl.BooleanExpression
    ) -> pddl.BooleanExpression | None:
        return self._apply(formula)

    def visit_falsity(
        self, formula: pddl.BooleanExpression
    ) -> pddl.BooleanExpression | None:
        return self._apply(formula)

    def visit_atom(
        self, formula: pddl.BooleanExpression
    ) -> pddl.BooleanExpression | None:
        return self._apply(formula)

    def visit_numeric_condition(
        self, formula: pddl.BooleanExpression
    ) -> pddl.BooleanExpression | None:
        return self._apply(formula)


def _type_predicate_name(type_name: str) -> str:
    return f"@type-{type_name}@"


TypeClosure = dict[str, list[str]]


def get_type_atoms(
    parameters: Iterable[pddl.ArgumentDefinition],
    arg_type: type[pddl.Argument] = pddl.VariableArgument,
) -> list[pddl.Atom]:
    return [
        pddl.Atom(_type_predicate_name(param.type_name), [arg_type(param.name)])
        for param in parameters
    ]


def get_type_atoms_full_closure(
    type_closure: TypeClosure,
    parameters: Iterable[pddl.ArgumentDefinition],
    arg_type: type[pddl.Argument] = pddl.VariableArgument,
) -> list[pddl.Atom]:
    prec = []
    for param in parameters:
        prec.extend((
            pddl.Atom(_type_predicate_name(t), [arg_type(param.name)])
            for t in type_closure[param.type_name]
        ))
    return prec


def drop_types(
    args: Iterable[pddl.ArgumentDefinition],
) -> list[pddl.ArgumentDefinition]:
    return [arg.__class__(arg.name, "object") for arg in args]


class QuantificationTypeReplacer(pddl.BooleanExpressionTransformer):
    def __init__(self, type_closure: TypeClosure):
        self.type_closure = type_closure

    def visit_quantification(self, formula: pddl.Quantification) -> pddl.Quantification:
        atoms = get_type_atoms(formula.parameters)
        if len(atoms) == 0:
            return formula.sub_formula
        atoms.append(formula.sub_formula)
        return formula.__class__(
            drop_types(formula.parameters), pddl.Conjunction(atoms)
        )


class UniversalEffectReplacer(pddl.ActionEffectTransformer):
    def __init__(self, type_closure: TypeClosure):
        self.type_closure = type_closure

    def visit_universal_effect(self, effect: pddl.UniversalEffect) -> pddl.ActionEffect:
        atoms = get_type_atoms(effect.parameters)
        if len(atoms) == 0:
            return effect.effect
        return pddl.UniversalEffect(
            drop_types(effect.parameters),
            pddl.ConditionalEffect(pddl.Conjunction(atoms), effect.effect),
        )


def compile_away_types_domain(
    domain: pddl.Domain, problem: pddl.Problem, type_closure: TypeClosure
) -> pddl.Domain:
    assert type_closure["object"] == ["object"]
    # closure = get_type_closure(domain)
    # closure["object"] = []
    domain.types = tuple()
    problem.initial = problem.initial + tuple(
        itertools.chain((
            get_type_atoms_full_closure(
                type_closure, domain.constants, pddl.ObjectArgument
            )
        ))
    )
    domain.constants = tuple(drop_types(domain.constants))
    domain.predicates = tuple(
        itertools.chain(
            (
                pddl.Predicate(p.name, drop_types(p.parameters))
                for p in domain.predicates
            ),
            (
                pddl.Predicate(
                    _type_predicate_name(t), [pddl.ArgumentDefinition("?x", "object")]
                )
                for t in type_closure
            ),
        )
    )
    domain.functions = tuple(
        pddl.Function(p.name, drop_types(p.parameters)) for p in domain.functions
    )
    actions = []
    q_replacer = QuantificationTypeReplacer(type_closure)
    e_replacer = UniversalEffectReplacer(type_closure)
    for action in domain.actions:
        parameters = drop_types(action.parameters)
        precondition = action.precondition.traverse(q_replacer)
        effect = transform_all_conditions_in_effect(
            action.effect.traverse(e_replacer), q_replacer
        )
        atoms = get_type_atoms(action.parameters)
        if len(atoms) > 0:
            atoms.append(action.precondition)
            precondition = pddl.Conjunction(atoms)
        actions.append(pddl.Action(action.name, parameters, precondition, effect))
    domain.actions = tuple(actions)
    derived_predicates = []
    for pred in domain.derived_predicates:
        parameters = drop_types(pred.predicate.parameters)
        condition = pred.condition.traverse(q_replacer)
        atoms = get_type_atoms(pred.predicate.parameters)
        if len(atoms) > 0:
            atoms.append(condition)
            condition = pddl.Conjunction(atoms)
        derived_predicates.append(
            pddl.DerivedPredicate(
                pddl.Predicate(pred.predicate.name, parameters), condition
            )
        )
    domain.derived_predicates = tuple(derived_predicates)


def compile_away_types_problem(
    problem: pddl.Problem, type_closure: TypeClosure
) -> pddl.Problem:
    problem.initial = problem.initial + tuple(
        get_type_atoms_full_closure(type_closure, problem.objects, pddl.ObjectArgument)
    )
    problem.objects = tuple(drop_types(problem.objects))
    problem.goal = problem.goal.traverse(QuantificationTypeReplacer(type_closure))


class QuantifierSplitting(pddl.BooleanExpressionTransformer):
    def visit_exists(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        sub = formula.sub_formula
        if isinstance(sub, pddl.Disjunction):
            return pddl.Disjunction([
                pddl.Exists(formula.parameters, phi).simplify().traverse(self)
                for phi in sub.sub_formulas
            ]).simplify()
        sub = sub.traverse(self)
        return (
            formula
            if sub is formula.sub_formula
            else pddl.Exists(formula.parameters, sub)
        )

    def visit_forall(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        sub = formula.sub_formula
        if isinstance(sub, pddl.Conjunction):
            return pddl.Conjunction([
                pddl.Forall(formula.parameters, phi).simplify().traverse(self)
                for phi in sub.sub_formulas
            ]).simplify()
        sub = sub.traverse(self)
        return (
            formula
            if sub is formula.sub_formula
            else pddl.Forall(formula.parameters, sub)
        )


class QuantifierElimination(pddl.BooleanExpressionTransformer):
    def __init__(self):
        self.derived_predicates = []

    def next_derived_predicate_name(self) -> str:
        return f"@exists-{len(self.derived_predicates)}@"

    def visit_exists(self, formula: pddl.Exists) -> pddl.BooleanExpression:
        transformed_child = formula.sub_formula.traverse(self)
        free_child_vars = transformed_child.get_free_variables()
        quantified_vars = set((p.name for p in formula.parameters))
        pred_params = list(sorted(set(free_child_vars) - quantified_vars))
        pred_name = self.next_derived_predicate_name()
        predicate = pddl.Predicate(
            pred_name, [pddl.ArgumentDefinition(x, "object") for x in pred_params]
        )
        condition = pddl.Exists(
            formula.parameters,
            transformed_child,
        )
        self.derived_predicates.append(pddl.DerivedPredicate(predicate, condition))
        return pddl.Atom(pred_name, [pddl.VariableArgument(x) for x in pred_params])

    def visit_forall(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        assert False, "Quantifier elimination required normalized PDDL"


class DisjunctionElimination(pddl.BooleanExpressionTransformer):
    derived_predicates = []
    disjuncts = [0]

    @staticmethod
    def reset():
        DisjunctionElimination.derived_predicates = []
        DisjunctionElimination.disjuncts = [0]

    def __init__(self, var_types: dict[str, str]):
        self.var_types: dict[str, str] = var_types

    def visit_quantification(self, formula: pddl.Quantification) -> pddl.Quantification:
        for arg in formula.parameters:
            self.var_types[arg.name] = arg.type_name
        new_formula = formula.__class__(
            formula.parameters, formula.sub_formula.traverse(self)
        )
        for arg in formula.parameters:
            del self.var_types[arg.name]
        return new_formula

    def next_derived_predicate_name(self) -> str:
        self.disjuncts[0] += 1
        return f"@disjunct-{self.disjuncts[0] - 1}@"

    def visit_disjunction(
        self, formula: pddl.BooleanExpression
    ) -> pddl.BooleanExpression:
        name = self.next_derived_predicate_name()
        # collect free variables of *all* disjuncts
        params = formula.get_free_variables()
        predicate = pddl.Predicate(
            name, [pddl.ArgumentDefinition(x, self.var_types[x]) for x in params]
        )
        for disjunct in formula.sub_formulas:
            condition = disjunct.traverse(self)
            self.derived_predicates.append(pddl.DerivedPredicate(predicate, condition))
        return pddl.Atom(name, [pddl.VariableArgument(x) for x in params])


class MoveQuantifiersOutward(pddl.BooleanExpressionTransformer):
    def visit_conjunction(self, formula: pddl.Conjunction) -> pddl.BooleanExpression:
        simplified = pddl.Conjunction(
            (f.traverse(self) for f in formula.sub_formulas)
        ).simplify()
        if not isinstance(simplified, pddl.Conjunction):
            return simplified
        params = []
        subs = []
        for sub in simplified.sub_formulas:
            if isinstance(sub, pddl.Quantification):
                assert isinstance(sub, pddl.Exists)
                assert all(
                    itertools.chain.from_iterable(
                        ((a.name != b.name for b in params) for a in sub.parameters)
                    )
                )
                params.extend(sub.parameters)
                subs.append(sub.sub_formula)
            else:
                subs.append(sub)
        new_conj = pddl.Conjunction(subs).simplify()
        if len(params) == 0:
            return new_conj
        return pddl.Exists(params, new_conj)


def _apply_transformation(
    domain: pddl.Domain,
    problem: pddl.Problem,
    transformation: pddl.BooleanExpressionTransformer,
):
    transform_all_conditions(domain, transformation)
    problem.goal = problem.goal.traverse(transformation)


def _apply_transformation_with_vars(
    domain: pddl.Domain,
    problem: pddl.Problem,
    transformer_class: type[pddl.BooleanExpressionTransformer],
):
    transform_all_conditions_with_vars(domain, transformer_class)
    transformation = transformer_class({})
    problem.goal = problem.goal.traverse(transformation)


class PositiveLiteralChecker(pddl.BooleanExpressionVisitor):
    def __init__(self):
        self.result: bool = True

    def visit_generic(self, formula: pddl.BooleanExpression):
        self.result = False

    def visit_truth(self, formula: pddl.BooleanExpression):
        pass

    def visit_numeric_condition(self, formula: pddl.BooleanExpression):
        pass

    def visit_atom(self, formula: pddl.BooleanExpression):
        pass


class LiteralChecker(PositiveLiteralChecker):
    def visit_negation(self, formula: pddl.Negation):
        c = PositiveLiteralChecker()
        formula.sub_formula.traverse(c)
        self.result = self.result and c.result


class NormalFormChecker(LiteralChecker):
    def visit_conjunction(self, formula: pddl.Conjunction):
        for y in formula.sub_formulas:
            c = LiteralChecker()
            y.traverse(c)
            self.result = self.result and c.result


class Simplifier(pddl.BooleanExpressionTransformer):
    def visit_generic(self, formula: pddl.BooleanExpression) -> pddl.BooleanExpression:
        return formula.simplify()

    def visit_negation(self, formula: pddl.Negation) -> pddl.BooleanExpression:
        return formula.simplify()

    def visit_bool_connector(
        self, formula: pddl.BooleanConnector
    ) -> pddl.BooleanExpression:
        return formula.simplify()

    def visit_quantification(
        self, formula: pddl.Quantification
    ) -> pddl.BooleanExpression:
        return formula.simplify()


T = TypeVar("T")


def _filter_inplace(predicate: Callable[[T], bool], elems: list[T]):
    i = 0
    for j, elem in enumerate(elems):
        if predicate(elem):
            if i != j:
                elems[i] = elem
            i += 1
    if i < len(elems):
        del elems[i:]


def drop_inapplicable_actions(domain: pddl.Domain):
    _filter_inplace(
        lambda action: not isinstance(action.precondition, pddl.Falsity), domain.actions
    )


def drop_inapplicable_predicates(domain: pddl.Domain):
    _filter_inplace(
        lambda pred: not isinstance(pred.condition, pddl.Falsity),
        domain.derived_predicates,
    )


def drop_inapplicable_effects(domain: pddl.Domain):
    class RemoveEffect(pddl.ActionEffectTransformer):
        def visit_conditional_effect(
            self, effect: pddl.ConditionalEffect
        ) -> pddl.ActionEffect | None:
            if isinstance(effect.condition, pddl.Falsity):
                return None
            return pddl.ConditionalEffect(
                effect.condition, effect.effect.traverse(self)
            )

        def visit_probabilistic_effect(
            self, effect: pddl.ProbabilisticEffect
        ) -> pddl.ActionEffect | None:
            new_outs = list(
                filter(
                    lambda out: out.effect is not None,
                    map(
                        lambda out: pddl.ProbabilisticOutcome(
                            out.probability, out.effect.traverse(self)
                        ),
                        effect.outcomes,
                    ),
                )
            )
            if len(new_outs) == 0:
                return None
            return pddl.ProbabilisticEffect(new_outs)

        def visit_universal_effect(
            self, effect: pddl.UniversalEffect
        ) -> pddl.ActionEffect | None:
            eff = effect.effect.traverse(self)
            if eff is None:
                return None
            return pddl.UniversalEffect(effect.parameters, eff)

        def visit_conjunctive_effect(
            self, effect: pddl.ConjunctiveEffect
        ) -> pddl.ActionEffect | None:
            new_effs = list(
                filter(
                    lambda eff: eff is not None,
                    map(lambda eff: eff.traverse(self), effect.effects),
                )
            )
            if len(new_effs) == 0:
                return None
            return pddl.ConjunctiveEffect(new_effs)

    for action in domain.actions:
        action.effect = action.effect.traverse(RemoveEffect())

    _filter_inplace(lambda action: action.effect is not None, domain.actions)


def normalize_conditions(domain: pddl.Domain, problem: pddl.Problem):
    # 0. rename all variables, making their names distinct
    standardize_variable_names(domain, problem)

    # 1. push negation inwards
    _apply_transformation(domain, problem, PushNegation())

    # 2. simplify
    _apply_transformation(domain, problem, Simplifier())

    # 3. move quantifiers inwards
    _apply_transformation(domain, problem, QuantifierSplitting())

    # 4. replace forall conditions by not(exists)
    _apply_transformation(domain, problem, ReplaceForall())

    # 1. push negation inwards
    _apply_transformation(domain, problem, PushNegation())
    _apply_transformation(domain, problem, Simplifier())

    # 5. replace quantifiers by derivied predicates
    qe = QuantifierElimination()
    _apply_transformation(domain, problem, qe)
    domain.derived_predicates = domain.derived_predicates + tuple(qe.derived_predicates)

    # simplify
    _apply_transformation(domain, problem, Simplifier())

    # 5. replace disjunctions by derived predicates
    _apply_transformation_with_vars(domain, problem, DisjunctionElimination)
    domain.derived_predicates = domain.derived_predicates + tuple(
        DisjunctionElimination.derived_predicates
    )
    DisjunctionElimination.reset()

    # 3. compile away types
    type_closure = get_type_closure(domain)
    compile_away_types_domain(domain, problem, type_closure)
    compile_away_types_problem(problem, type_closure)

    # renormalize quantifiers (in derived predicates)
    _apply_transformation(domain, problem, MoveQuantifiersOutward())

    # 6. simplify all conditions
    _apply_transformation(domain, problem, Simplifier())

    # drop trivially inapplicable actions / effects
    drop_inapplicable_predicates(domain)
    drop_inapplicable_actions(domain)
    drop_inapplicable_effects(domain)

    # collections functions
    fns = FunctionsCollector()
    visit_all_expressions(domain, fns)
    if problem.metric is not None:
        problem.metric.expression.traverse(fns)
    builtins = [f for f in fns.functions if f in ("total-cost", "reward")]
    for f in builtins:
        if not any((ff.name == f for ff in domain.functions)):
            domain.functions = (*domain.functions, pddl.Function(f, []))

    # ====
    # now all conditions should be conjunctions of literals
    nnf = NormalFormChecker()
    for action in domain.actions:
        action.precondition.traverse(nnf)
        visit_all_conditions_in_effect(action.effect, nnf)
    problem.goal.traverse(nnf)
    assert nnf.result


class PushConditionalInwards(pddl.ActionEffectTransformer):
    def visit_conditional_effect(
        self, effect: pddl.ConditionalEffect
    ) -> pddl.ActionEffect:
        class Pusher(pddl.ActionEffectTransformer):
            def __init__(self, condition: pddl.BooleanExpression):
                self.condition: pddl.BooleanExpression = condition

            def visit_generic(self, effect: pddl.ActionEffect) -> pddl.ActionEffect:
                return pddl.ConditionalEffect(self.condition, effect)

            def visit_negative_effect(
                self, effect: pddl.NegativeEffect
            ) -> pddl.ActionEffect:
                return pddl.ConditionalEffect(self.condition, effect)

            def visit_conditional_effect(
                self, effect: pddl.ConditionalEffect
            ) -> pddl.ActionEffect:
                extended = Pusher(
                    pddl.Conjunction([self.condition, effect.condition]).simplify()
                )
                return effect.effect.traverse(extended)

            def visit_universal_effect(
                self, effect: pddl.UniversalEffect
            ) -> pddl.ActionEffect:
                return pddl.UniversalEffect(
                    effect.parameters, effect.effect.traverse(self)
                )

            def visit_probabilistic_effect(
                self, effect: pddl.ProbabilisticEffect
            ) -> pddl.ActionEffect:
                outcomes = []
                for outcome in effect.outcomes:
                    outcomes.append(
                        pddl.ProbabilisticOutcome(
                            outcome.probability, outcome.effect.traverse(self)
                        )
                    )
                return pddl.ProbabilisticEffect(outcomes)

        pusher = Pusher(effect.condition)
        return effect.effect.traverse(pusher)


class MoveUniversalIn(pddl.ActionEffectTransformer):
    def __init__(self, parameters: Iterable[pddl.Argument]):
        self.parameters: list[pddl.ArgumentDefinition] = list(parameters)

    def visit_universal_effect(self, effect: pddl.UniversalEffect) -> pddl.ActionEffect:
        vis = MoveUniversalIn(self.parameters)
        vis.parameters.extend(effect.parameters)
        return effect.effect.traverse(vis)

    def visit_generic(self, effect: pddl.ActionEffect) -> pddl.ActionEffect:
        if len(self.parameters) > 0:
            free_vars = effect.get_free_variables()
            params = [p for p in self.parameters if p.name in free_vars]
            if len(params) > 0:
                return pddl.UniversalEffect(params, effect)
        return effect

    def visit_conditional_effect(
        self, effect: pddl.ConditionalEffect
    ) -> pddl.ActionEffect:
        return self.visit_generic(effect)

    def visit_probabilistic_effect(
        self, effect: pddl.ProbabilisticEffect
    ) -> pddl.ActionEffect:
        return self.visit_generic(effect)

    # def _make_effect(self, effect: pddl.ActionEffect) -> pddl.ActionEffect:
    #     if not self.is_root or len(self.parameters) == 0:
    #         return effect
    #     return pddl.UniversalEffect(self.parameters, effect)
    #
    # def visit_universal_effect(self, eff: pddl.UniversalEffect) -> pddl.ActionEffect:
    #     self.parameters.extend(eff.parameters)
    #     root, self.is_root = self.is_root, False
    #     new_eff = eff.effect.traverse(self)
    #     self.is_root = root
    #     return self._make_effect(new_eff)
    #
    # def visit_conjunctive_effect(
    #     self, eff: pddl.ConjunctiveEffect
    # ) -> pddl.ActionEffect:
    #     root, self.is_root = self.is_root, False
    #     new_eff = super().visit_conjunctive_effect(eff)
    #     self.is_root = root
    #     return self._make_effect(new_eff)
    #
    # def visit_probabilistic_effect(
    #     self, effect: pddl.ProbabilisticEffect
    # ) -> pddl.ActionEffect:
    #     root, self.is_root = self.is_root, False
    #     new_eff = super().visit_probabilistic_effect(effect)
    #     self.is_root = root
    #     return self._make_effect(new_eff)
    #
    # def visit_conditional_effect(
    #     self, effect: pddl.ConditionalEffect
    # ) -> pddl.ActionEffect:
    #     root, self.is_root = self.is_root, False
    #     new_eff = super().visit_conditional_effect(effect)
    #     self.is_root = root
    #     return self._make_effect(new_eff)


class MovePEffOut(pddl.ActionEffectTransformer):
    def visit_conjunctive_effect(
        self, effect: pddl.ConjunctiveEffect
    ) -> pddl.ActionEffect:
        prob_effs = []
        univ_effs = []
        other_effs = []
        for eff in effect.effects:
            efff = eff.traverse(self)
            if isinstance(efff, pddl.ProbabilisticEffect):
                prob_effs.append(efff)
            elif isinstance(efff, pddl.UniversalEffect):
                univ_effs.append(efff)
            else:
                other_effs.append(efff)

        if len(prob_effs) * len(other_effs) == 0:
            return effect

        new_effects = univ_effs
        for prob_eff in prob_effs:
            total_prob = Fraction(0, 1)
            expr = pddl.NumericConstant("1")
            non_const_prob = False
            outcomes = []
            for outcome in prob_eff.outcomes:
                outcomes.append(
                    pddl.ProbabilisticOutcome(
                        outcome.probability,
                        pddl.ConjunctiveEffect(
                            [*other_effs, outcome.effect]
                        ).simplify(),
                    )
                )
                expr = pddl.Subtraction(expr, outcome.probability)
                if isinstance(outcome.probability, pddl.NumericConstant):
                    total_prob += Fraction(outcome.probability.value_str)
                else:
                    non_const_prob = True
            if non_const_prob:
                outcomes.append(
                    pddl.ProbabilisticOutcome(
                        expr, pddl.ConjunctiveEffect(other_effs).simplify()
                    )
                )
            elif total_prob < Fraction(1):
                prob = Fraction(1) - total_prob
                outcomes.append(
                    pddl.ProbabilisticOutcome(
                        pddl.NumericConstant(str(prob)),
                        pddl.ConjunctiveEffect(other_effs).simplify(),
                    )
                )
            new_effects.append(pddl.ProbabilisticEffect(outcomes))

        if len(new_effects) == 1:
            return new_effects[0]
        return pddl.ConjunctiveEffect(new_effects)


def _transform_effects(domain: pddl.Domain, transformer: pddl.ActionEffectTransformer):
    for action in domain.actions:
        action.effect = action.effect.traverse(transformer)


def normalize_effects(domain: pddl.Domain):
    # 1. push conditional effects inwards
    _transform_effects(domain, PushConditionalInwards())

    # 2. move universal out
    _transform_effects(domain, MoveUniversalIn([]))

    # 3. move atomic effects into probabilistic effects
    _transform_effects(domain, MovePEffOut())

    # 3. simplify (i.e., collapse conjunctions etc)
    for action in domain.actions:
        action.effect = action.effect.simplify()

    # => all effects are have hierarchy
    # Conjunction[Universal[Probabilistic]]
    # Universal > Conjunction > Probabilistic > Conjunction > Universal > Conditional >
    # Atomic



import pytest

import plado.parser.tokenizer
import plado.pddl as pddl
from plado.parser.parser import (
    LookaheadStreamer,
    _parse_condition,
    _parse_effect,
    _parse_expression,
    parse_action,
    parse_constants,
    parse_derived_predicate,
    parse_domain,
    parse_functions,
    parse_initial,
    parse_metric,
    parse_predicates,
    parse_problem,
)


def tokenize(text):
    return LookaheadStreamer(plado.parser.tokenizer.tokenize(text))


@pytest.fixture
def objects():
    num = 10
    return " ".join(["o%d" % i for i in range(num)]) + " - object)"


@pytest.fixture
def predicates():
    return """
        (p)
        (q ?x)
        (r ?x ?y)
    )
    """


@pytest.fixture
def typed_predicates():
    return """
        (p ?x - type0 ?y ?z - type1 ?zz)
    )
    """


@pytest.fixture
def typed_functions():
    return "(f) - number (g ?x ?y ?z - sometype))"


def test_streamer_consume_all(objects):
    streamer = tokenize(objects)
    n = 0
    for tok in streamer:
        n += 1
    assert n == 13


def test_streamer_lookahead(objects):
    streamer = tokenize(objects)
    assert streamer.at(5).tok == "o5"
    n = 0
    for tok in streamer:
        n += 1
    assert n == 13


def test_parse_objects(objects):
    parsed = parse_constants(tokenize(objects))
    assert len(parsed) == 10
    assert all((o.type_name == "object" for o in parsed))
    assert all((o.name == "o%d" % i for i, o in enumerate(parsed)))


def test_parse_predicates(predicates):
    parsed = parse_predicates(tokenize(predicates))
    assert len(parsed) == 3
    assert parsed[0].name == "p" and len(parsed[0].parameters) == 0
    assert parsed[1].name == "q" and len(parsed[1].parameters) == 1
    assert parsed[2].name == "r" and len(parsed[2].parameters) == 2


def test_parse_typed_predicates(typed_predicates):
    parsed = parse_predicates(tokenize(typed_predicates))
    assert len(parsed) == 1
    assert parsed[0].name == "p" and len(parsed[0].parameters) == 4
    assert (
        parsed[0].parameters[0].name == "?x"
        and parsed[0].parameters[0].type_name == "type0"
    )
    assert (
        parsed[0].parameters[1].name == "?y"
        and parsed[0].parameters[1].type_name == "type1"
    )
    assert (
        parsed[0].parameters[2].name == "?z"
        and parsed[0].parameters[2].type_name == "type1"
    )
    assert (
        parsed[0].parameters[3].name == "?zz"
        and parsed[0].parameters[3].type_name == "object"
    )


def test_parse_typed_functions(typed_functions):
    parsed = parse_functions(tokenize(typed_functions))
    assert len(parsed) == 2
    assert parsed[0].name == "f" and len(parsed[0].parameters) == 0
    assert (
        parsed[1].name == "g"
        and len(parsed[1].parameters) == 3
        and all((t.type_name == "sometype" for t in parsed[1].parameters))
    )


def test_parse_empty_condition():
    parsed = _parse_condition((tokenize("()")))
    assert isinstance(parsed, pddl.Truth)


def test_parse_atom_conditions():
    parsed = _parse_condition((tokenize("(p)")))
    assert (
        isinstance(parsed, pddl.Atom)
        and parsed.name == "p"
        and len(parsed.arguments) == 0
    )


def test_parse_negation():
    formula = "(not (p ?x))"
    parsed = _parse_condition(tokenize(formula))
    assert isinstance(parsed, pddl.Negation)
    assert isinstance(parsed.sub_formula, pddl.Atom)


def test_parse_atom_parameterized_conditions():
    parsed = _parse_condition((tokenize("(p ?x ?y ?z )")))
    assert (
        isinstance(parsed, pddl.Atom)
        and parsed.name == "p"
        and len(parsed.arguments) == 3
    )


def test_parse_empty_conjunctive_condition():
    parsed = _parse_condition((tokenize("(and)")))
    assert isinstance(parsed, pddl.Conjunction) and len(parsed.sub_formulas) == 0


def test_parse_singleton_conjunction_condition():
    parsed = _parse_condition((tokenize("(and (p))")))
    assert isinstance(parsed, pddl.Conjunction) and len(parsed.sub_formulas) == 1
    assert isinstance(parsed.sub_formulas[0], pddl.Atom)


def test_parse_nested_conjunction():
    formula = "(and (or (p) (q)) (or (p) (q)))"
    parsed = _parse_condition(tokenize(formula))
    assert isinstance(parsed, pddl.Conjunction) and len(parsed.sub_formulas) == 2
    assert (
        isinstance(parsed.sub_formulas[0], pddl.Disjunction)
        and len(parsed.sub_formulas[0].sub_formulas) == 2
    )
    assert (
        isinstance(parsed.sub_formulas[1], pddl.Disjunction)
        and len(parsed.sub_formulas[1].sub_formulas) == 2
    )


def test_parse_forall_condition():
    formula = "(forall (?x ?y - t) (and))"
    parsed = _parse_condition(tokenize(formula))
    assert (
        isinstance(parsed, pddl.Forall)
        and len(parsed.parameters) == 2
        and isinstance(parsed.sub_formula, pddl.Conjunction)
    )


def test_parse_implication_condition():
    formula = "(imply (and) (or (p)))"
    parsed = _parse_condition(tokenize(formula))
    assert isinstance(parsed, pddl.Disjunction) and len(parsed.sub_formulas) == 2
    assert isinstance(parsed.sub_formulas[0], pddl.Negation)
    assert isinstance(parsed.sub_formulas[1], pddl.Disjunction)


def test_parse_numeric_condition():
    formula = "(< (f) 100)"
    parsed = _parse_condition(tokenize(formula))
    assert isinstance(parsed, pddl.Less)
    assert isinstance(parsed.lhs, pddl.FunctionCall)
    assert isinstance(parsed.rhs, pddl.NumericConstant)


def test_parse_numeric_condition2():
    formula = "(= (f ?x ?y) (+ (f) (g)))"
    parsed = _parse_condition(tokenize(formula))
    assert isinstance(parsed, pddl.Equals)
    assert isinstance(parsed.lhs, pddl.FunctionCall)
    assert isinstance(parsed.rhs, pddl.Sum)


def test_parse_atom_effect():
    formula = "(p)"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.AtomEffect) and parsed.name == "p"


def test_parse_conjunctive_effect():
    formula = "(and (p) (q))"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.ConjunctiveEffect) and len(parsed.effects) == 2
    assert isinstance(parsed.effects[0], pddl.AtomEffect)
    assert isinstance(parsed.effects[1], pddl.AtomEffect)


def test_parse_conditional_effect():
    formula = "(when (p) (q))"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.ConditionalEffect)
    assert isinstance(parsed.condition, pddl.Atom)
    assert isinstance(parsed.effect, pddl.AtomEffect)


def test_universal_effect():
    formula = "(forall (?x) (when (p ?x) (q ?x)))"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.UniversalEffect)
    assert len(parsed.parameters) == 1
    assert isinstance(parsed.effect, pddl.ConditionalEffect)


def test_probabilistic_effect():
    formula = "(probabilistic 0.5 (p) 1/2 (q))"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.ProbabilisticEffect) and len(parsed.outcomes) == 2
    assert isinstance(parsed.outcomes[0], pddl.ProbabilisticOutcome)
    assert isinstance(parsed.outcomes[1], pddl.ProbabilisticOutcome)


def test_nested_probabilistic_effect():
    formula = "(probabilistic 0.5 (when (p) (probabilistic 0.5 (q))))"
    parsed = _parse_effect(tokenize(formula))
    assert isinstance(parsed, pddl.ProbabilisticEffect) and len(parsed.outcomes) == 1
    effect = parsed.outcomes[0]
    assert (
        isinstance(effect, pddl.ProbabilisticOutcome)
        and isinstance(effect.effect, pddl.ConditionalEffect)
        and isinstance(effect.effect.effect, pddl.ProbabilisticEffect)
    )


def test_parse_initial_state():
    formula = "(p) (= (f) 1))"
    parsed = parse_initial(tokenize(formula))
    assert len(parsed) == 2
    assert isinstance(parsed[0], pddl.Atom)
    assert (
        isinstance(parsed[1], pddl.NumericAssignEffect)
        and parsed[1].variable.name == "f"
    )


def test_parse_metric():
    formula = "minimize (total-cost))"
    parsed = parse_metric(tokenize(formula))
    assert isinstance(parsed, pddl.Metric)
    assert parsed.direction == pddl.Metric.MINIMIZE
    assert isinstance(parsed.expression, pddl.FunctionCall)
    assert parsed.expression.name == "total-cost"


@pytest.fixture
def simple_domain():
    domain = """
(define (domain test-1234)
(:types t)
(:predicates (p ?x - t))
(:functions (total-cost) - number)
(:derived (p ?x - t) (forall (?y - t) (not (p ?y))))
(:action set
    :parameters (?x - t)
    :precondition ()
    :effect (p ?x))
    )
    """
    return parse_domain(tokenize(domain))


def test_simple_domain_name(simple_domain):
    assert simple_domain.name == "test-1234"


def test_simple_domain_types(simple_domain):
    assert len(simple_domain.types) == 1
    t = next((x for x in simple_domain.types if x.name == "t"), [None])
    assert t is not None
    assert t.parent_type_name == "object"


def test_simple_domain_predicates(simple_domain):
    assert len(simple_domain.predicates) == 1
    p = next((x for x in simple_domain.predicates if x.name == "p"), [None])
    assert p is not None and len(p.parameters) == 1 and p.parameters[0].type_name == "t"


def test_simple_domain_functions(simple_domain):
    assert len(simple_domain.functions) == 1
    total_cost = next(
        (x for x in simple_domain.functions if x.name == "total-cost"), [None]
    )
    assert total_cost is not None


def test_simple_domain_derived_predicates(simple_domain):
    assert len(simple_domain.derived_predicates) == 1
    dp = simple_domain.derived_predicates[0]
    assert (
        dp.predicate.name == "p"
        and len(dp.predicate.parameters) == 1
        and dp.predicate.parameters[0].name == "?x"
        and dp.predicate.parameters[0].type_name == "t"
    )
    assert isinstance(dp.condition, pddl.Forall)
    assert (
        len(dp.condition.parameters) == 1
        and dp.condition.parameters[0].name == "?y"
        and dp.condition.parameters[0].type_name == "t"
    )
    neg = dp.condition.sub_formula
    assert isinstance(neg, pddl.Negation)
    atom = neg.sub_formula
    assert isinstance(atom, pddl.Atom)
    assert (
        atom.name == "p" and len(atom.arguments) == 1 and atom.arguments[0].name == "?y"
    )


def test_simple_domain_actions(simple_domain):
    assert len(simple_domain.actions) == 1
    action = simple_domain.actions[0]
    assert action.name == "set"
    assert len(action.parameters) == 1
    assert action.parameters[0].name == "?x" and action.parameters[0].type_name == "t"
    assert isinstance(action.precondition, pddl.Truth)
    assert (
        isinstance(action.effect, pddl.AtomEffect)
        and action.effect.name == "p"
        and len(action.effect.arguments) == 1
        and action.effect.arguments[0].name == "?x"
    )


@pytest.fixture
def simple_problem():
    problem = """
    (define (problem test-problem)
    (:domain test-domain)
    (:objects o1 o2 o3 - object)
    (:init (p o1))
    (:goal (and (p o2) (p o3)))
    (:metric minimize (total-cost))
    )
    """
    return parse_problem(tokenize(problem))


def test_simple_problem_name(simple_problem):
    assert simple_problem.name == "test-problem"


def test_simple_problem_domain_name(simple_problem):
    assert simple_problem.domain_name == "test-domain"


def test_simple_problem_objects(simple_problem):
    assert len(simple_problem.objects) == 3
    assert all((
        any((o.name == "o%i" % i for o in simple_problem.objects)) for i in range(1, 4)
    ))


def test_simple_problem_init(simple_problem):
    assert len(simple_problem.initial) == 1
    p = simple_problem.initial[0]
    assert (
        isinstance(p, pddl.Atom)
        and p.name == "p"
        and len(p.arguments) == 1
        and p.arguments[0].name == "o1"
    )


def test_simple_problem_goal(simple_problem):
    assert isinstance(simple_problem.goal, pddl.Conjunction)
    assert len(simple_problem.goal.sub_formulas) == 2


def test_simple_problem_goal_reward(simple_problem):
    assert simple_problem.goal_reward is None


def test_simple_problem_metric(simple_problem):
    assert isinstance(simple_problem.metric, pddl.Metric)
    assert simple_problem.metric.direction == pddl.Metric.MINIMIZE
    assert (
        isinstance(simple_problem.metric.expression, pddl.FunctionCall)
        and simple_problem.metric.expression.name == "total-cost"
    )



from collections.abc import Iterable
from typing import Any

from plado.pddl.arguments import Argument, ArgumentDefinition, Substitution
from plado.pddl.boolean_expression import BooleanExpression, Conjunction, Falsity, Truth
from plado.pddl.numeric_expression import FunctionCall, NumericExpression, Product

CR = "\n"


class ActionEffect:
    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        raise NotImplementedError()

    def substitute(self, subs: Substitution) -> "ActionEffect":
        return self

    def simplify(self) -> "ActionEffect":
        raise NotImplementedError()

    def get_free_variables(self) -> set[str]:
        return NotImplementedError()

    def dump(self, level: int) -> str:
        raise NotImplementedError()


class AtomEffect(ActionEffect):
    def __init__(self, name: str, arguments: Iterable[Argument]):
        self.name: str = name
        self.arguments: tuple[Argument] = tuple(arguments)

    def substitute(self, subs: Substitution) -> "AtomEffect":
        return AtomEffect(self.name, (arg.apply(subs) for arg in self.arguments))

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_atom_effect(self)

    def __str__(self):
        return self.dump(0)

    def simplify(self) -> "AtomEffect":
        return self

    def get_free_variables(self) -> set[str]:
        return set((arg.name for arg in self.arguments if arg.is_variable()))

    def dump(self, level: int) -> str:
        return (
            f"({self.name}"
            f"{' ' if len(self.arguments) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.arguments))}"
            ")"
        )


class NegativeEffect(ActionEffect):
    def __init__(self, atom: AtomEffect):
        self.atom: AtomEffect = atom

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_negative_effect(self)

    def substitute(self, subs: Substitution) -> "NegativeEffect":
        return NegativeEffect(self.atom.substitute(subs))

    def simplify(self) -> "NegativeEffect":
        return self

    def get_free_variables(self) -> set[str]:
        return self.atom.get_free_variables()

    def dump(self, level: int) -> str:
        return f"(not {self.atom.dump(level)})"


class ConjunctiveEffect(ActionEffect):
    def __init__(self, effects: Iterable[ActionEffect]):
        self.effects: tuple[ActionEffect] = tuple(effects)

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_conjunctive_effect(self)

    def substitute(self, subs: Substitution) -> "ConjunctiveEffect":
        return ConjunctiveEffect((eff.substitute(subs) for eff in self.effects))

    def simplify(self) -> "ConjunctiveEffect":
        merged = False
        new_effs = []
        for eff in self.effects:
            if isinstance(eff, ConjunctiveEffect):
                new_effs.extend(eff.effects)
                merged = True
            else:
                new_effs.append(eff)
        return ConjunctiveEffect(new_effs) if merged else self

    def get_free_variables(self) -> set[str]:
        variables = set()
        for eff in self.effects:
            variables = variables | eff.get_free_variables()
        return variables

    def dump(self, level: int) -> str:
        if len(self.effects) == 0:
            return "(and )"
        if len(self.effects) == 1:
            return f"(and {self.effects[0].dump(level+1)})"
        return "".join([
            "(and\n",
            CR.join((" " * (level * 2 + 2) + e.dump(level + 1) for e in self.effects)),
            CR,
            " " * (level * 2),
            ")",
        ])


class ConditionalEffect(ActionEffect):
    def __init__(self, condition: BooleanExpression, effect: ActionEffect):
        self.condition: BooleanExpression = condition
        self.effect: ActionEffect = effect

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_conditional_effect(self)

    def substitute(self, subs: Substitution) -> "ConditionalEffect":
        return ConditionalEffect(
            self.condition.substitute(subs), self.effect.substitute(subs)
        )

    def simplify(self) -> ActionEffect:
        cond = self.condition.simplify()
        if isinstance(cond, Truth):
            return self.effect.simplify()
        if isinstance(cond, Falsity):
            return ConjunctiveEffect([])
        eff = self.effect.simplify()
        if isinstance(eff, ConjunctiveEffect) and len(eff.effects) == 0:
            return eff
        if isinstance(eff, ConditionalEffect):
            return ConditionalEffect(
                Conjunction((cond, eff.condition)), eff.effect
            ).simplify()
        return ConditionalEffect(cond, eff)

    def get_free_variables(self) -> set[str]:
        return self.condition.get_free_variables() | self.effect.get_free_variables()

    def dump(self, level: int) -> str:
        return (
            "(when\n"
            f"{' ' * (level * 2)}  {self.condition.dump(level + 1)}\n"
            f"{' ' * (level * 2)}  {self.effect.dump(level + 1)})"
        )


class UniversalEffect(ActionEffect):
    def __init__(self, parameters: Iterable[ArgumentDefinition], effect: ActionEffect):
        self.parameters = tuple(parameters)
        self.effect = effect

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_universal_effect(self)

    def substitute(self, subs: Substitution) -> "UniversalEffect":
        new_subs = {
            x: y
            for (x, y) in subs.items()
            if all((p.name != x.name) for p in self.parameters)
        }
        return UniversalEffect(self.parameters, self.effect.substitute(new_subs))

    def get_free_variables(self) -> set[str]:
        params = set((p.name for p in self.parameters))
        return self.effect.get_free_variables() - params

    def simplify(self) -> ActionEffect:
        eff = self.effect.simplify()
        if isinstance(eff, ConjunctiveEffect) and len(eff.effects) == 0:
            return eff
        if isinstance(eff, UniversalEffect):
            return UniversalEffect(
                self.parameters + eff.parameters, eff.effect
            ).simplify()
        variables = eff.get_free_variables()
        new_params = [p for p in self.parameters if p.name in variables]
        if len(new_params) == 0:
            return eff
        return UniversalEffect(new_params, eff)

    def dump(self, level: int) -> str:
        return (
            f"(forall ({' '.join((str(arg) for arg in self.parameters))})\n"
            f"{' ' * (level * 2 + 2)}{self.effect.dump(level+1)})"
        )


class ProbabilisticOutcome:
    def __init__(self, probability: NumericExpression, effect: ActionEffect):
        self.probability: NumericExpression = probability
        self.effect: ActionEffect = effect

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_probabilistic_effect(self)

    def substitute(self, subs: Substitution) -> "ProbabilisticOutcome":
        return ProbabilisticOutcome(
            self.probability.substitute(subs), self.effect.substitute(subs)
        )

    def get_free_variables(self) -> set[str]:
        return self.effect.get_free_variables() | self.probability.get_free_variables()

    def simplify(self) -> "ProbabilisticOutcome":
        return ProbabilisticOutcome(self.probability, self.effect.simplify())

    def dump(self, level: int) -> str:
        return f"{self.probability.dump(level)} {self.effect.dump(level)}"


class ProbabilisticEffect(ActionEffect):
    def __init__(self, outcomes: Iterable[ProbabilisticOutcome]):
        self.outcomes: tuple[ProbabilisticOutcome] = tuple(outcomes)

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_probabilistic_effect(self)

    def substitute(self, subs: Substitution) -> "ProbabilisticEffect":
        return ProbabilisticEffect((e.substitute(subs) for e in self.outcomes))

    def get_free_variables(self) -> set[str]:
        variables = set()
        for out in self.outcomes:
            variables = variables | out.get_free_variables()
        return variables

    def simplify(self) -> ActionEffect:
        new_outs = []
        for out in self.outcomes:
            new_out = out.simplify()
            if isinstance(new_out.effect, ProbabilisticEffect):
                new_outs.extend((
                    ProbabilisticOutcome(
                        Product(out.probability, x.probability),
                        x.effect,
                    )
                    for x in new_out.effect.outcomes
                ))
            elif (
                not isinstance(new_out.effect, ConjunctiveEffect)
                or len(new_out.effect.effects) > 0
            ):
                new_outs.append(new_out)
        if len(new_outs) == 0:
            return ConjunctiveEffect([])
        return ProbabilisticEffect(new_outs)

    def dump(self, level: int) -> str:
        return "".join([
            "(probabilistic\n",
            CR.join((" " * (level * 2 + 2) + o.dump(level + 1) for o in self.outcomes)),
            "\n",
            (" " * (level * 2)),
            ")",
        ])


class NumericEffect(ActionEffect):
    OP = None

    def __init__(self, variable: FunctionCall, expression: NumericExpression):
        self.variable: FunctionCall = variable
        self.expression: NumericExpression = expression

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        raise NotImplementedError()

    def substitute(self, subs: Substitution) -> "NumericEffect":
        return self.__class__(
            self.variable.substitute(subs), self.expression.substitute(subs)
        )

    def get_free_variables(self) -> set[str]:
        return self.variable.get_free_variables() | self.expression.get_free_variables()

    def simplify(self) -> "NumericEffect":
        return self

    def dump(self, level: int) -> str:
        return f"({self.OP} {self.variable.dump(level)} {self.expression.dump(level)})"


class NumericAssignEffect(NumericEffect):
    OP = "="

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_assign_effect(self)


class ScaleUpEffect(NumericEffect):
    OP = "scale-up"

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_scale_up_effect(self)


class ScaleDownEffect(NumericEffect):
    OP = "scale-down"

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_scale_down_effect(self)


class IncreaseEffect(NumericEffect):
    OP = "increase"

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_increase_effect(self)


class DecreaseEffect(NumericEffect):
    OP = "decrease"

    def traverse(self, visitor: "ActionEffectVisitor") -> Any:
        return visitor.visit_decrease_effect(self)


class ActionEffectVisitor:
    def visit_generic(self, effect: ActionEffect) -> Any:
        raise NotImplementedError()

    def visit_atom_effect(self, effect: AtomEffect) -> Any:
        return self.visit_generic(effect)

    def visit_negative_effect(self, effect: NegativeEffect) -> Any:
        return self.visit_generic(effect)

    def visit_conjunctive_effect(self, effect: ConjunctiveEffect) -> Any:
        return self.visit_generic(effect)

    def visit_conditional_effect(self, effect: ConditionalEffect) -> Any:
        return self.visit_generic(effect)

    def visit_universal_effect(self, effect: UniversalEffect) -> Any:
        return self.visit_generic(effect)

    def visit_probabilistic_effect(self, effect: ProbabilisticEffect) -> Any:
        return self.visit_generic(effect)

    def visit_numeric_effect(self, effect: NumericEffect) -> Any:
        return self.visit_generic(effect)

    def visit_assign_effect(self, effect: NumericAssignEffect) -> Any:
        return self.visit_numeric_effect(effect)

    def visit_scale_up_effect(self, effect: ScaleUpEffect) -> Any:
        return self.visit_numeric_effect(effect)

    def visit_scale_down_effect(self, effect: ScaleDownEffect) -> Any:
        return self.visit_numeric_effect(effect)

    def visit_increase_effect(self, effect: IncreaseEffect) -> Any:
        return self.visit_numeric_effect(effect)

    def visit_decrease_effect(self, effect: DecreaseEffect) -> Any:
        return self.visit_numeric_effect(effect)


class RecursiveActionEffectVisitor(ActionEffectVisitor):
    def visit_generic(self, effect: ActionEffect):
        assert False, "visit_generic should have been unreachable"

    def visit_atomic(self, effect: ActionEffect) -> Any:
        raise NotImplementedError()

    def visit_atom_effect(self, effect: AtomEffect) -> Any:
        return self.visit_atomic(effect)

    def visit_numeric_effect(self, effect: NumericEffect) -> Any:
        return self.visit_atomic(effect)

    def visit_negative_effect(self, effect: NegativeEffect) -> Any:
        return self.visit_atom_effect(effect.atom)

    def visit_conjunctive_effect(self, effect: ConjunctiveEffect) -> Any:
        for e in effect.effects:
            if e.traverse(self):
                return True
        return False

    def visit_conditional_effect(self, effect: ConditionalEffect) -> Any:
        return effect.effect.traverse(self)

    def visit_universal_effect(self, effect: UniversalEffect) -> Any:
        return effect.effect.traverse(self)

    def visit_probabilistic_effect(self, effect: ProbabilisticEffect) -> Any:
        for e in effect.outcomes:
            if e.effect.traverse(self):
                return True
        return False


class ActionEffectTransformer(ActionEffectVisitor):
    def visit_generic(self, effect: ActionEffect) -> ActionEffect:
        return effect

    def visit_negative_effect(self, effect: NegativeEffect) -> ActionEffect:
        return NegativeEffect(effect.atom.traverse(self))

    def visit_conjunctive_effect(self, effect: ConjunctiveEffect) -> ActionEffect:
        return ConjunctiveEffect((e.traverse(self) for e in effect.effects))

    def visit_conditional_effect(self, effect: ConditionalEffect) -> ActionEffect:
        return ConditionalEffect(effect.condition, effect.effect.traverse(self))

    def visit_universal_effect(self, effect: UniversalEffect) -> ActionEffect:
        return UniversalEffect(effect.parameters, effect.effect.traverse(self))

    def visit_probabilistic_effect(self, effect: ProbabilisticEffect) -> ActionEffect:
        return ProbabilisticEffect((
            ProbabilisticOutcome(e.probability, e.effect.traverse(self))
            for e in effect.outcomes
        ))



from plado import pddl


def transform_all_conditions_in_effect(
    eff: pddl.ActionEffect, transformer: pddl.BooleanExpressionTransformer
) -> pddl.ActionEffect:
    class CondTransformer(pddl.ActionEffectTransformer):
        def visit_conditional_effect(
            self, eff: pddl.ConditionalEffect
        ) -> pddl.ConditionalEffect:
            cond = eff.condition.traverse(transformer)
            return pddl.ConditionalEffect(cond, eff.effect.traverse(self))

    ct = CondTransformer()
    return eff.traverse(ct)


def transform_all_conditions(
    domain: pddl.Domain, transformer: pddl.BooleanExpressionTransformer
):
    for action in domain.actions:
        action.precondition = action.precondition.traverse(transformer)
        action.effect = transform_all_conditions_in_effect(action.effect, transformer)
    for predicate in domain.derived_predicates:
        predicate.condition = predicate.condition.traverse(transformer)


def transform_all_conditions_in_effect_with_vars(
    eff: pddl.ActionEffect,
    var_types: dict[str, str],
    transformer_class: type[pddl.BooleanExpressionTransformer],
) -> pddl.ActionEffect:
    class CondTransformer(pddl.ActionEffectTransformer):
        def __init__(self, var_types: dict[str, str]):
            self.var_types: dict[str, str] = var_types

        def visit_universal_effect(
            self, effect: pddl.UniversalEffect
        ) -> pddl.UniversalEffect:
            for p in effect.parameters:
                self.var_types[p.name] = p.type_name
            new_eff = pddl.UniversalEffect(
                effect.parameters, effect.effect.traverse(self)
            )
            for p in effect.parameters:
                del self.var_types[p.name]
            return new_eff

        def visit_conditional_effect(
            self, effect: pddl.ConditionalEffect
        ) -> pddl.ConditionalEffect:
            transformer = transformer_class(self.var_types)
            cond = effect.condition.traverse(transformer)
            return pddl.ConditionalEffect(cond, effect.effect.traverse(self))

    ct = CondTransformer(var_types)
    return eff.traverse(ct)


def transform_all_conditions_with_vars(
    domain: pddl.Domain, transformer_class: type[pddl.BooleanExpressionTransformer]
):
    for action in domain.actions:
        transformer = transformer_class(
            {p.name: p.type_name for p in action.parameters}
        )
        action.precondition = action.precondition.traverse(transformer)
        action.effect = transform_all_conditions_in_effect_with_vars(
            action.effect,
            {p.name: p.type_name for p in action.parameters},
            transformer_class,
        )
    for predicate in domain.derived_predicates:
        transformer = transformer_class(
            {p.name: p.type_name for p in predicate.predicate.parameters}
        )
        predicate.condition = predicate.condition.traverse(transformer)



#!/bin/env/python

import argparse
import sys

from plado.parser import LookaheadStreamer, parse_domain, parse_problem, tokenize
from plado.pddl.effects import ActionEffectTransformer


class NumericEffectRemover(ActionEffectTransformer):
    def visit_numeric_effect(self, effect) -> None:
        return None

    def visit_universal_effect(self, effect):
        result = super().visit_universal_effect(effect)
        if result.effect is None:
            return None
        return result

    def visit_probabilistic_effect(self, effect):
        outcomes = []
        for outcome in effect.outcomes:
            e = outcome.effect.traverse(self)
            if e is not None:
                outcomes.append(outcome.__class__(outcome.probability, e))
        if len(outcomes) == 0:
            return None
        return effect.__class_(outcomes)

    def visit_conditional_effect(self, effect):
        result = super().visit_conditional_effect(effect)
        if result.effect is None:
            return None
        return result

    def visit_conjunctive_effect(self, effect):
        result = super().visit_conjunctive_effect(effect)
        result.effects = tuple(eff for eff in result.effects if eff is not None)
        if len(result.effects) == 0:
            return None
        return result


def remove_action_costs_from_domain(domain_path: str) -> str:
    with open(domain_path, encoding="ascii") as f:
        domain = parse_domain(LookaheadStreamer(tokenize(f.read())))
    domain.requirements = tuple(r for r in domain.requirements if r != ":action-costs")
    actions = []
    for action in domain.actions:
        effect = action.effect.traverse(NumericEffectRemover())
        if effect is not None:
            action.effect = effect
            actions.append(action)
    domain.actions = tuple(actions)
    domain.functions = tuple()
    return domain.dump()


def remove_action_costs_from_problem(problem_path: str) -> str:
    with open(problem_path, encoding="ascii") as f:
        problem = parse_problem(LookaheadStreamer(tokenize(f.read())))
    i = [f for f in problem.initial if not hasattr(f, "expression")]
    problem.initial = tuple(i)
    problem.metric = None
    return problem.dump()


def main():
    p = argparse.ArgumentParser()
    p.add_argument("file_path", help="Path to PDDL domain or problem file")
    p.add_argument(
        "--inplace",
        "-i",
        help=(
            "Overwrite content of the given file. Otherwise result is printed to"
            " console."
        ),
        action="store_true",
    )
    args = p.parse_args()
    try:
        dom = remove_action_costs_from_domain(args.file_path)
        if args.inplace:
            with open(args.file_path, "w", encoding="ascii") as f:
                f.write(dom)
        else:
            print(dom)
        return 0
    except Exception:
        pass
    try:
        prob = remove_action_costs_from_problem(args.file_path)
        if args.inplace:
            with open(args.file_path, "w", encoding="ascii") as f:
                f.write(prob)
        else:
            print(prob)
        return 0
    except Exception:
        pass
    print("failed parsing PDDL", file=sys.stderr)
    return 1


if __name__ == "__main__":
    main()



from plado.datalog.evaluator import Atom, Clause, DatalogEngine, DatalogProgram
from plado.semantics.task import State, Task


class GoalChecker:
    def __init__(self, task: Task):
        self.task: Task = task
        program: DatalogProgram = task.create_datalog_program()
        self.gr: int = program.num_relations()
        program.add_relation(0)
        gc = Clause(Atom(self.gr, []), [], [], [])
        task.goal.condition.to_datalog(len(task.predicates), gc)
        program.add_clause(gc)
        self.evalutor: DatalogEngine = DatalogEngine(program, len(task.objects))

    def __call__(self, state: State) -> bool:
        atoms, fluents = self.task.prepare_for_query(state.atoms, state.fluents)
        atoms.append(set())
        model = self.evalutor(atoms, fluents)
        return len(model[self.gr]) > 0
